import numpy as np


class ProfilePlotModel():

    def get_plot_data(self, data, crosshair, region,
                      phi_sample=180, line_sample=50):

        if region == 'x':
            x_points = line_sample * [crosshair.x]
            y_points = np.linspace(
                data.y_axis[0], data.y_axis[-1], num=line_sample,
                endpoint=True)

        elif region == 'y':
            y_points = line_sample * [crosshair.y]
            x_points = np.linspace(
                data.x_axis[0], data.x_axis[-1], num=line_sample,
                endpoint=True)

        elif region == 'roi' or region == 'border' or region == 'ring':
            # Circular parametrization in mathematical direction
            angles = np.linspace(0, 2 * np.pi,
                                 num=phi_sample, endpoint=False)

            # Find radii smaple points
            if region == 'roi':
                radii = np.linspace(0, crosshair.radius,
                                    num=line_sample, endpoint=True)

            elif region == 'border':
                radii = np.array(crosshair.radius)

            else:
                # Ring
                radii = np.linspace(crosshair.radius,
                                    crosshair.radius + crosshair.width,
                                    num=line_sample,
                                    endpoint=True)

            x_points = []
            y_points = []
            for phi in angles:
                x_points.append(radii * np.cos(phi) + crosshair.x)
                y_points.append(radii * np.sin(phi) + crosshair.y)

        x_points = np.array(x_points).flatten()
        y_points = np.array(y_points).flatten()
        intensities = np.array(data.interpolate_points(
            x_points, y_points, fill_value=np.nan))

        if region == 'x':
            x = y_points
            y = intensities

        elif region == 'y':
            x = x_points
            y = intensities

        elif region == 'border':
            x = np.degrees(angles)
            y = intensities

        else:
            x = np.degrees(angles)
            intensities = intensities.reshape(phi_sample, line_sample)
            y = np.nanmean(intensities, 1)

        x, y = self._filter_nan(np.array(x), np.array(y))

        return x, y

    def _filter_nan(self, x, y):

        mask = np.ones(x.shape, dtype=bool)

        mask[np.isnan(x)] = False
        mask[np.isnan(y)] = False

        return x[mask], y[mask]



# Python Imports
import logging
import traceback
import datetime

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import pyqtSignal, QDir

# Own Imports
from kmap import __directory__
from kmap.controller.sliceddatatab import SlicedDataTab
from kmap.controller.orbitaldatatab import OrbitalDataTab
from kmap.controller.profileplottab import ProfilePlotTab
from kmap.controller.renametabwindow import RenameTabWindow
from kmap.controller.lmfitplottab import LMFitPlotTab
from kmap.controller.lmfittab import LMFitTab, LMFitResultTab
from kmap.controller.filetab import FileViewerTab, FileEditorTab
from kmap.library.qwidgetsub import Tab
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/tabwidget.ui')
TabWidget_UI, _ = uic.loadUiType(UI_file)


class TabWidget(QWidget, TabWidget_UI):

    tab_added = pyqtSignal(Tab)

    def __init__(self, *args, **kwargs):

        # Setup GUI
        super(TabWidget, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def open_sliced_data_tab(self, path):
        # Opens a new sliced data tab

        log = logging.getLogger('kmap')
        log.info('Trying to load %s' % path)

        try:
            tab = SlicedDataTab.init_from_path(path)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)

        except Exception as e:

            log.error('Couldn\'t load %s' % path)
            log.error(traceback.format_exc())

    def open_sliced_data_tab_by_URLs(self, URLs):

        try:
            tab = SlicedDataTab.init_from_URLs(URLs)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)

        except ValueError as e:
            pass

        except Exception as e:

            log = logging.getLogger('kmap')
            log.error('Couldn\'t load URLs')
            log.error(traceback.format_exc())

    def open_sliced_data_tab_by_URL(self, URL):

        try:
            tab = SlicedDataTab.init_from_URL(URL)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)
            
        except ValueError as e:
            pass

        except Exception as e:

            log = logging.getLogger('kmap')
            log.error('Couldn\'t load URL')
            log.error(traceback.format_exc())

    def open_sliced_data_tab_by_cube(self, URL):

        try:
            tab = SlicedDataTab.init_from_cube(URL)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)
            
        except ValueError as e:
            pass

        except Exception as e:

            log = logging.getLogger('kmap')
            log.error('Couldn\'t load cube file')
            log.error(traceback.format_exc())

    def open_orbital_data_tab(self):
        # Opens a new orbital data tab

        tab = OrbitalDataTab()
        title = tab.get_title()

        self._open_tab(tab, title)

        return tab

    def open_file_tab(self, path, title, editable=False, richText=False):

        if editable:
            tab = FileEditorTab(path, title)
        else:
            tab = FileViewerTab(path, title, richText=richText)

        self._open_tab(tab, title)

    def open_lmfit_tab(self, sliced_tab, orbital_tab):

        tab = LMFitTab(sliced_tab.get_data(), orbital_tab.get_orbitals())
        tab.fit_finished.connect(self.open_result_tab)

        self._open_tab(tab, 'LM-Fit Tab')

    def open_result_tab(self, *args):

        tab = LMFitResultTab(*args)
        tab.open_plot_tab.connect(self.open_lmfit_plot_tab)

        current_time = datetime.datetime.now()
        title = 'Results (%i:%i)' % (current_time.hour, current_time.minute)
        tab.set_title(title)
        self._open_tab(tab, title)

    def open_lmfit_plot_tab(self, results, orbitals, axis):

        tab = LMFitPlotTab(results, orbitals, axis)
        title = 'Plot'
        tab.set_title(title)
        self._open_tab(tab, title)

    def open_profile_tab(self):

        tab = ProfilePlotTab(self)

        self._open_tab(tab, 'Profile Plot')

    def get_orbital_tab_to_load_to(self):

        tab = None
        aux = self.tab_widget.currentWidget()
        if type(aux) == OrbitalDataTab:
            tab = aux

        else:
            for index in range(self.tab_widget.count()):
                aux = self.tab_widget.widget(index)

                if type(aux) == OrbitalDataTab:
                    tab = aux
                    break

        if tab is None:
            tab = self.open_orbital_data_tab()

        return tab

    def get_current_tab(self):

        return self.tab_widget.currentWidget()

    def get_tabs_of_type(self, type_):

        tabs = []

        for index in range(self.tab_widget.count()):
            tab = self.tab_widget.widget(index)

            if type(tab) == type_ or type_ is None:
                tabs.append(tab)

        return tabs

    def rename_current_tab(self):

        self.rename_tab = RenameTabWindow()
        self.rename_tab.title_chosen.connect(self.set_tab_title)

    def set_tab_title(self, title):

        current_tab = self.get_current_tab()
        current_tab_index = self.tab_widget.indexOf(current_tab)

        current_tab.set_title(title)
        self.tab_widget.setTabText(current_tab_index, title)

        for tab in self.get_tabs_of_type(ProfilePlotTab):
            tab.rename_loaded_tab(current_tab, title)

    def close_tab(self, index):
        # Close tab specified with index

        widget = self.tab_widget.widget(index)
        self.tab_widget.removeTab(index)
        widget.close()
        widget.deleteLater()

    def _open_tab(self, tab, title, tooltip=None):

        index = self.tab_widget.addTab(tab, title)
        self.tab_widget.setCurrentIndex(index)

        if tooltip is not None:
            self.tab_widget.setTabToolTip(index, tooltip)

        self.tab_added.emit(tab)

    def _connect(self):

        self.tab_widget.tabCloseRequested.connect(self.close_tab)
        self.tab_widget.tabBarDoubleClicked.connect(self.rename_current_tab)



from kmap.config.config import config
from kmap.library.misc import axis_from_range


class MatplotlibWindowModel():

    def __init__(self, plot_data):

        self.image = plot_data.data
        self.x, self.y = self._calc_centered_axes(plot_data)

    def _calc_centered_axes(self, plot_data):

        x_step_size = plot_data.step_size[0]
        centered_range = plot_data.range[0] + \
            [-x_step_size / 2, x_step_size / 2]
        x = axis_from_range(centered_range, len(plot_data.x_axis) + 1)

        y_step_size = plot_data.step_size[1]
        centered_range = plot_data.range[1] + \
            [-y_step_size / 2, y_step_size / 2]
        y = axis_from_range(centered_range, len(plot_data.y_axis) + 1)

        return x, y



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal, QDir
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/cubeoptions.ui')
CubeOptions_UI, _ = uic.loadUiType(UI_file)


class CubeOptions(QWidget, CubeOptions_UI):

    energy_changed = pyqtSignal()
    resolution_changed = pyqtSignal()
    symmetrization_changed = pyqtSignal()

    def __init__(self, *args, **kwargs):

        # Setup GUI
        super(CubeOptions, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def get_parameters(self):

        energy = self.energy_spinbox.value()
        resolution = self.resolution_spinbox.value()
        symmetry = self._get_symmetry()

        return energy, resolution, symmetry

    def _get_symmetry(self):

        index = self.symmetrize_combobox.currentIndex()
        available_symmetries = ['no', '2-fold', '2-fold+mirror',
                                '3-fold', '3-fold+mirror', '4-fold',
                                '4-fold+mirror']

        return available_symmetries[index]

    def _connect(self):

        self.energy_spinbox.valueChanged.connect(self.energy_changed.emit)
        self.resolution_spinbox.valueChanged.connect(
            self.resolution_changed.emit)
        self.symmetrize_combobox.currentIndexChanged.connect(
            self.symmetrization_changed.emit)



from kmap.library.id import ID
from kmap.library.sliceddata import SlicedData
from kmap.library.misc import transpose_axis_order

class SlicedDataTabModel():

    def __init__(self):

        self.data = None

    def load_data_from_URLs(self, URLs):  # -> create data[BE,kx,ky]

        # Last element in URLs are the parameters. All other elements
        # are individual orbitals to load. Each is a list of length 2
        # with first entry being the URL, the second the meta_data
        # dictionary
        *orbitals, options = URLs
        name, *parameters = options
        self.data = SlicedData.init_from_orbitals(name, orbitals, parameters)

        self.change_slice(0, 0)

    def load_data_from_URL(self, URL):  # -> create data[photon_energy,kx,ky]

        # Last element in URL are the parameters. The first element
        # is the orbital to load as a list of length 2
        # with first entry being the URL, the second the meta_data
        # dictionary
        *orbital, options = URL
        name, *parameters = options
        self.data = SlicedData.init_from_orbital_photonenergy(
            name, orbital, parameters)

        self.change_slice(0, 0)

    def load_data_from_cube(self, URL):  # -> create either psi[x,y,z]
                                         #               or psik[kx,ky,kz]

        # Last element in URL are the parameters. The first element
        # is the orbital to load as a list of length 2
        # with first entry being the URL, the second the meta_data
        # dictionary
        *orbital, options = URL
        name, *parameters = options
        self.data = SlicedData.init_from_orbital_cube(
            name, orbital, parameters)

        self.change_slice(0, 0)
        
    def transpose(self, constant_axis):

        axis_order = transpose_axis_order(constant_axis)

        self.data.transpose(axis_order)

    def load_data_from_path(self, path):

        self.data = SlicedData.init_from_hdf5(path)

        self.change_slice(0, 0)

    def change_slice(self, index, axis):

        self.displayed_plot_data = self.data.slice_from_index(index, axis)

        return self.displayed_plot_data

    def to_string(self):

        rep = '%s' % str(self.data)

        return rep



# Python Imports
import logging
import urllib.request

# Third Party Imports
import h5py
import numpy as np

# Own Imports
from kmap.library.id import ID
from kmap.library.plotdata import PlotData
from kmap.library.abstractdata import AbstractData
from kmap.library.misc import axis_from_range, energy_to_k
from kmap.library.database import Database
from kmap.library.orbital import Orbital
from kmap.library.axis import Axis


class SlicedData(AbstractData):

    def __init__(self, name, axis_1, axis_2, axis_3, data, meta_data={}):

        if isinstance(name, str) and name:
            super(SlicedData, self).__init__(ID.new_ID(), name, meta_data)

        else:
            raise ValueError('name has to be string and not empty')

        data = np.array(data, dtype=np.float64)
        if len(data.shape) == 3:
            self.data = data

        else:
            raise ValueError('data has to be 3D')

        axis_1 = Axis.init_from_hdf_list(axis_1, len(data[:, 0, 0]))
        axis_2 = Axis.init_from_hdf_list(axis_2, len(data[0, :, 0]))
        axis_3 = Axis.init_from_hdf_list(axis_3, len(data[0, 0, :]))
        self.axes = [axis_1, axis_2, axis_3]

    @classmethod
    def init_from_hdf5(cls, file_path, keys={}, meta_data={}):

        # Updates default file_keys with user defined keys
        file_keys = {'name': 'name', 'axis_1_label': 'axis_1_label',
                     'axis_1_units': 'axis_1_units',
                     'axis_1_range': 'axis_1_range',
                     'axis_2_label': 'axis_2_label',
                     'axis_2_units': 'axis_2_units',
                     'axis_2_range': 'axis_2_range',
                     'axis_3_label': 'axis_3_label',
                     'axis_3_units': 'axis_3_units',
                     'axis_3_range': 'axis_3_range',
                     'data': 'data'}
        file_keys.update(keys)

        with h5py.File(file_path, 'r') as file:

            # First check if necessary datasets exist
            for _, value in file_keys.items():
                if value not in file:
                    raise AttributeError('Dataset is missing %s' % value)

            # Read all datasets
            for key, value in file.items():
                if key == file_keys['name']:
                    name = str(file[key][()])

                elif key == file_keys['axis_1_label']:
                    axis_1_label = file[key][()]

                elif key == file_keys['axis_2_label']:
                    axis_2_label = file[key][()]

                elif key == file_keys['axis_3_label']:
                    axis_3_label = file[key][()]

                elif key == file_keys['axis_1_units']:
                    axis_1_units = file[key][()]

                elif key == file_keys['axis_2_units']:
                    axis_2_units = file[key][()]

                elif key == file_keys['axis_3_units']:
                    axis_3_units = file[key][()]

                elif key == file_keys['axis_1_range']:
                    axis_1_range = file[key][()]

                elif key == file_keys['axis_2_range']:
                    axis_2_range = file[key][()]

                elif key == file_keys['axis_3_range']:
                    axis_3_range = file[key][()]

                elif key == file_keys['data']:
                    data = file[key][()]

                else:
                    meta_data.update({key: str(file[key][()])})

        axis_1 = [axis_1_label, axis_1_units, axis_1_range]
        axis_2 = [axis_2_label, axis_2_units, axis_2_range]
        axis_3 = [axis_3_label, axis_3_units, axis_3_range]

        return cls(name, axis_1, axis_2, axis_3, data, meta_data)

    @classmethod
    def init_from_orbitals(cls, name, orbitals, parameters):
        """Returns a SlicedData object with the data[BE,kx,ky] 
           computed from the kmaps of several orbitals and
           broadened in energy.

        Args:
            name (str): name for SlicedData object
            orbitals (list): [[ 'URL1',dict1], ['URL2',dict2], ...]
                dict needs keys: 'energy' and 'name' 
            parameters (list): list of parameters
                photon_energy (float): Photon energy in eV.
                fermi_energy (float): Fermi energy in eV
                energy_broadening (float): FWHM of Gaussian energy broadenening in eV
                dk (float): Desired k-resolution in kmap in Angstroem^-1.
                phi (float): Euler orientation angle phi in degree.
                theta (float): Euler orientation angle theta in degree.
                psi (float): Euler orientation angle psi in degree.
                Ak_type (string): Treatment of |A.k|^2: either 'no',
                        'toroid' or 'NanoESCA'.
                polarization (string): Either 'p', 's', 'unpolarized', C+', 'C-' or
                        'CDAD'.   
                alpha (float): Angle of incidence plane in degree.
                beta (float): Azimuth of incidence plane in degree.
                gamma (float/str): Damping factor for final state in
                        Angstroem^-1. str = 'auto' sets gamma automatically
                symmetrization (str): either 'no', '2-fold', '2-fold+mirror',
                        '3-fold', '3-fold+mirror','4-fold', '4-fold+mirror'
        Returns:
            (SlicedData): SlicedData containing kmaps of all orbitals
        """

        log = logging.getLogger('kmap')
        # extract parameters
        photon_energy = parameters[0]
        fermi_energy = parameters[1]
        energy_broadening = parameters[2]
        dk = parameters[3]
        phi, theta, psi = parameters[4], parameters[5], parameters[6]
        Ak_type = parameters[7]
        polarization = parameters[8]
        alpha, beta, gamma = parameters[9], parameters[10], parameters[11]
        symmetrization = parameters[12]

        # determine axis_1 from minimal and maximal binding energy
        extend_range = 3 * energy_broadening
        energies = []

        for orbital in orbitals:
            energies.append(orbital[1]['energy'])

        BE_min = min(energies) - fermi_energy - extend_range
        BE_max = max(energies) - fermi_energy + extend_range
        # set energy grid spacing 6 times smaller than broadening
        dBE = energy_broadening / 6
        nBE = int((BE_max - BE_min) / dBE) + 1
        BE = np.linspace(BE_min, BE_max, nBE)
        nBE = len(BE)
        axis_1 = ['-BE', 'eV', [BE_min, BE_max]]

        # determine axis_2 and axis_3 kmax at BE_max
        Phi = -fermi_energy  # work function
        E_kin_max = photon_energy - Phi + BE_max
        k_max = energy_to_k(E_kin_max)
        nk = int((2 * k_max) / dk) + 1
        k_grid = np.linspace(-k_max, +k_max, nk)
        nk = len(k_grid)
        axis_2 = ['kx', '1/Å', [-k_max, +k_max]]
        axis_3 = ['ky', '1/Å', [-k_max, +k_max]]

        # initialize 3D-numpy array with zeros
        data = np.zeros((nBE, nk, nk))
        orbital_names = []

        # add kmaps of orbitals to
        log.info('Adding orbitals to SlicedData Object, please wait!')
        for orbital in orbitals:
            # binding energy of orbital
            BE0 = orbital[1]['energy'] - fermi_energy
            # kinetic energy of emitted electron
            E_kin = photon_energy - Phi + BE0

            # Gaussian weight function
            norm = (1 / np.sqrt(2 * np.pi * energy_broadening**2))
            weight = norm * \
                np.exp(-((BE - BE0)**2 / (2 * energy_broadening**2)))

            url = orbital[0]
            log.info('Loading from database: %s' % url)
            with urllib.request.urlopen(url) as f:
                file = f.read().decode('utf-8')
                orbital_data = Orbital(file)

            orbital_names.append(orbital[1]['name'])
            log.info('Computing k-map for %s' % orbital[1]['name'])

            kmap = orbital_data.get_kmap(E_kin, dk, phi, theta, psi,
                                         Ak_type, polarization, alpha, beta,
                                         gamma, symmetrization)
            kmap.interpolate(k_grid, k_grid, update=True)
            log.info('Adding to 3D-array: %s' % orbital[1]['name'])
            for i in range(len(BE)):
                data[i, :, :] += weight[i] * np.nan_to_num(kmap.data)

        # set NaNs outside photoemission horizon
        KX, KY = np.meshgrid(k_grid, k_grid)
        for i in range(len(BE)):
            E_kin = photon_energy - Phi + BE[i]
            k_max = energy_to_k(E_kin)
            out = np.sqrt(KX**2 + KY**2) > k_max
            tmp = data[i, :, :]
            tmp[out] = np.NaN
            data[i, :, :] = tmp

        # define meta-data for tool-tip display
        orbital_info = {}
        for orbital_name, energy in zip(orbital_names, energies):
            orbital_info[orbital_name] = energy

        meta_data = {'Photon energy (eV)': photon_energy,
                     'Fermi energy (eV)': fermi_energy,
                     'Energy broadening (eV)': energy_broadening,
                     'Molecular orientation': (phi, theta, psi),
                     '|A.k|^2 factor': Ak_type,
                     'Polarization': polarization,
                     'Incidence direction': (alpha, beta),
                     'Symmetrization': symmetrization,
                     'Orbital Info': orbital_info}

        return cls(name, axis_1, axis_2, axis_3, data, meta_data)

    @classmethod
    def init_from_orbital_cube(cls, name, orbital, parameters):
        """Returns a SlicedData object with the data[x,y,z] 
           taken from the real space wave function in orbital.

        Args:
            name (str): name for SlicedData object
            orbital (list): [ 'URL',dict ]
                dict needs keys: 'energy' and 'name'         
            parameters (list): list of parameters
                domain (str): either 'real-space' or 'k-space'
                dk3D (float): Desired resolution for 3D-Fourier-Transform.
                    Single number.
                E_kin_max (float): maximum kinetic energy in eV is used to
                    reduce the size of the 3D-numpy-array in momentum space
                value (string): choose between 'real', 'imag', 'abs' or 'abs2'
                    for Re(), Im(), |..| or |..|^2

        Returns:
            (SlicedData): SlicedData containing the real space orbital
        """
        log = logging.getLogger('kmap')

        orbital = orbital[0]  # only consider first orbital in list!

        # extract parameters
        domain = parameters[0]
        dk3D = parameters[1]
        E_kin_max = parameters[2]
        value = parameters[3]

        orbital_file = orbital[0]
        log.info('Loading from database: %s' % orbital_file)

        # decision if reading cube-file from URL or local file
        if orbital_file[:4] == 'http':
            with urllib.request.urlopen(orbital_file) as f:
                file = f.read().decode('utf-8')

        else:
            file = open(orbital_file).read()

        orbital_data = Orbital(
            file, dk3D=dk3D, E_kin_max=E_kin_max, value=value)

        # set name for SliceData object
        name = orbital_data.psi['name']

        # set axis and data
        if domain == 'real-space':
            axis_1 = ['x', 'Å', [orbital_data.psi['x']
                                 [0], orbital_data.psi['x'][-1]]]
            axis_2 = ['y', 'Å', [orbital_data.psi['y']
                                 [0], orbital_data.psi['y'][-1]]]
            axis_3 = ['z', 'Å', [orbital_data.psi['z']
                                 [0], orbital_data.psi['z'][-1]]]
            data = orbital_data.psi['data']

        else:
            axis_1 = ['kx', '1/Å', [orbital_data.psik['kx']
                                    [0], orbital_data.psik['kx'][-1]]]
            axis_2 = ['ky', '1/Å', [orbital_data.psik['ky']
                                    [0], orbital_data.psik['ky'][-1]]]
            axis_3 = ['kz', '1/Å', [orbital_data.psik['kz']
                                    [0], orbital_data.psik['kz'][-1]]]
            data = orbital_data.psik['data']

        # no meta data
        meta_data = {}

        return cls(name, axis_1, axis_2, axis_3, data, meta_data)

    @classmethod
    def init_from_orbital_photonenergy(cls, name, orbital, parameters):
        """Returns a SlicedData object with the data[photonenergy,kx,ky] 
           computed from the kmaps of one orbital for a series of
           photon energies.

        Args:
            name (str): name for SlicedData object
            orbital (list): [ 'URL',dict ]
                dict needs keys: 'energy' and 'name' 
            parameters (list): list of parameters
                hnu_min (float): minimal photon energy
                hnu_max (float): maximal photon energy
                hnu_step (float): stepsize for photon energy
                fermi_energy (float): Fermi energy in eV
                dk (float): Desired k-resolution in kmap in Angstroem^-1.
                phi (float): Euler orientation angle phi in degree.
                theta (float): Euler orientation angle theta in degree.
                psi (float): Euler orientation angle psi in degree.
                Ak_type (string): Treatment of |A.k|^2: either 'no',
                        'toroid' or 'NanoESCA'.
                polarization (string): Either 'p', 's', 'unpolarized', C+', 'C-' or
                        'CDAD'.   
                alpha (float): Angle of incidence plane in degree.
                beta (float): Azimuth of incidence plane in degree.
                gamma (float/str): Damping factor for final state in
                        Angstroem^-1. str = 'auto' sets gamma automatically
                symmetrization (str): either 'no', '2-fold', '2-fold+mirror',
                        '3-fold', '3-fold+mirror','4-fold', '4-fold+mirror'
        Returns:
            (SlicedData): SlicedData containing kmaps for various photon energies
        """

        log = logging.getLogger('kmap')
        orbital = orbital[0]  # only consider first orbital in list!

        # extract parameters
        hnu_min = parameters[0]
        hnu_max = parameters[1]
        hnu_step = parameters[2]
        fermi_energy = parameters[3]
        dk = parameters[4]
        phi, theta, psi = parameters[5], parameters[6], parameters[7]
        Ak_type = parameters[8]
        polarization = parameters[9]
        alpha, beta, gamma = parameters[10], parameters[11], parameters[12]
        symmetrization = parameters[13]

        # binding energy of orbital and work function
        BE = orbital[1]['energy'] - fermi_energy
        Phi = -fermi_energy  # work function

        # determine axis_1 = photon energy
        hnu = np.arange(hnu_min, hnu_max, hnu_step)
        n_hnu = len(hnu)

        axis_1 = ['photonenergy', 'eV', [hnu_min, hnu_max]]

        # determine axis_2 and axis_3 kmax at BE_max
        E_kin_max = hnu[-1] - Phi + BE
        k_max = energy_to_k(E_kin_max)
        nk = int((2 * k_max) / dk) + 1
        k_grid = np.linspace(-k_max, +k_max, nk)
        nk = len(k_grid)
        axis_2 = ['kx', '1/Å', [-k_max, +k_max]]
        axis_3 = ['ky', '1/Å', [-k_max, +k_max]]

        # initialize 3D-numpy array with zeros
        data = np.zeros((n_hnu, nk, nk))
        orbital_names = []

        log.info('Adding orbital to SlicedData Object, please wait!')
        # read orbital from cube-file database
        url = orbital[0]
        log.info('Loading from database: %s' % url)
        with urllib.request.urlopen(url) as f:
            file = f.read().decode('utf-8')
            orbital_data = Orbital(file)

        # loop over photon energies
        for i in range(len(hnu)):

            # kinetic energy of emitted electron
            E_kin = hnu[i] - Phi + BE
            kmap = orbital_data.get_kmap(E_kin, dk, phi, theta, psi,
                                         Ak_type, polarization, alpha, beta,
                                         gamma, symmetrization)
            kmap.interpolate(k_grid, k_grid, update=True)
            data[i, :, :] = kmap.data

        # define meta-data for tool-tip display
        orbital_info = orbital[1]

        meta_data = {'Fermi energy (eV)': fermi_energy,
                     'Molecular orientation': (phi, theta, psi),
                     '|A.k|^2 factor': Ak_type,
                     'Polarization': polarization,
                     'Incidence direction': (alpha, beta),
                     'Symmetrization': symmetrization,
                     'Orbital Info': orbital_info}

        return cls(name, axis_1, axis_2, axis_3, data, meta_data)

    def transpose(self, axes_order):

        self.data = self.data.transpose(axes_order)
        self.axes = [self.axes[i] for i in axes_order]

    def slice_from_index(self, index, axis=0):

        if axis == 0:
            data = self.data[index, :, :]
            range_ = [self.axes[2].range, self.axes[1].range]

        elif axis == 1:
            data = self.data[:, index, :]
            range_ = [self.axes[2].range, self.axes[0].range]

        elif axis == 2:
            data = self.data[:, :, index]
            range_ = [self.axes[1].range, self.axes[0].range]

        else:
            raise ValueError('axis has to be between 1 and 3')

        return PlotData(data, range_)

    def __str__(self):

        rep = AbstractData.__str__(self)

        for index, axis in enumerate(self.axes):
            rep += '\n\nAxis %i\n%s' % (index, str(self.axes[index]))

        rep += '\n\n'
        return rep[:-2]



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import QDir, pyqtSignal

# Own Imports
from kmap import __directory__
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/lmfitoptions.ui')
LMFitOptions_UI, _ = uic.loadUiType(UI_file)


class LMFitOptions(QWidget, LMFitOptions_UI):

    fit_triggered = pyqtSignal()
    region_changed = pyqtSignal(str, bool)
    background_changed = pyqtSignal(str)
    method_changed = pyqtSignal(str)
    slice_policy_changed = pyqtSignal(str)

    equations_path = __directory__+ QDir.toNativeSeparators('/resources/misc/background_equations')

    def __init__(self, parent):

        # Setup GUI
        super(LMFitOptions, self).__init__()
        self.setupUi(self)
        self._setup()
        self._connect(parent)

    def get_region(self):

        text = self.region_comboBox.currentText()

        if text == 'Entire kMap':
            region = 'all'
            inverted = False

        elif text == 'Only ROI':
            region = 'roi'
            inverted = False

        elif text == 'Only Annulus':
            region = 'ring'
            inverted = False

        elif text == 'Except ROI':
            region = 'roi'
            inverted = True

        elif text == 'Except Annulus':
            region = 'ring'
            inverted = True

        return region, inverted

    def get_method(self):

        text = self.method_combobox.currentText()

        return text[text.find("(") + 1:text.find(")")]

    def get_slice_policy(self):

        index = self.slice_combobox.currentIndex()

        if index == 0:
            return 'only one'

        elif index == 1:
            return 'all'

        else:
            return 'all combined'

    def get_background(self):

        return self.background_combobox.currentText()

    def update_fit_button(self):

        self.fit_button.setText('Fit')
        self.fit_button.repaint()

    def _trigger_fit(self):

        self.fit_button.setText('Running')
        self.fit_button.repaint()

        self.fit_triggered.emit()

    def _change_region(self):

        region, inverted = self.get_region()

        self.region_changed.emit(region, inverted)

    def _change_method(self):

        method = self.get_method()

        self.method_changed.emit(method)

    def _change_slice_policy(self):

        slice_policy = self.get_slice_policy()

        self.slice_policy_changed.emit(slice_policy)

    def _change_background(self):

        equation = self.get_background()
        self.background_changed.emit(equation)

    def _setup(self):

        with open(LMFitOptions.equations_path, 'r') as file:
            equations = file.read().split('\n')

        for equation in equations:

            self.background_combobox.addItem(equation)

        self.background_combobox.setCurrentIndex(0)

    def _connect(self, parent):

        parent.fit_finished.connect(self.update_fit_button)
        self.fit_button.clicked.connect(self._trigger_fit)
        self.region_comboBox.currentIndexChanged.connect(
            self._change_region)
        self.method_combobox.currentIndexChanged.connect(self._change_method)
        self.slice_combobox.currentIndexChanged.connect(
            self._change_slice_policy)
        self.background_combobox.currentIndexChanged.connect(
            self._change_background)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal, QDir
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/dataslider.ui')
DataSlider_UI, _ = uic.loadUiType(UI_file)


class DataSlider(QWidget, DataSlider_UI):

    slice_changed = pyqtSignal(int)
    axis_changed = pyqtSignal(int)
    tranpose_triggered = pyqtSignal(int)

    def __init__(self, data, *args, **kwargs):

        self.data = data

        # Setup GUI
        super(DataSlider, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()
        self._connect()

    def change_slice(self, index):

        sender = self.sender().objectName()

        # If spinbox or slider called, index will be the slice index
        # and the other widget has to be updated. As default use slider
        # position as index
        if sender == 'spinbox':
            self._update_slider_silently(index)

        elif sender == 'slider':
            self._update_spinbox_silently(index)

        else:
            index = self.slider.sliderPosition()

        self._update_slice_label()

        self.slice_changed.emit(index)

    def change_axis(self, axis):

        index = self.slider.sliderPosition()

        self._update_slider_silently(index)
        self._update_spinbox_silently(index)
        self._update_slice_label()

        self.axis_changed.emit(axis)

    def get_index(self):

        index = self.slider.sliderPosition()

        return index

    def get_axis(self):

        axis = self.combobox.currentIndex()

        return axis

    def trigger_transpose(self):

        index = self.slider.sliderPosition()

        self.tranpose_triggered.emit(self.get_axis())

        for i in range(3):
            self.combobox.setItemText(i, self.data.axes[i].label)

        self._update_slice_label()
        self._update_slider_silently(index)
        self._update_spinbox_silently(index)

    def _update_slice_label(self):

        index = self.slider.sliderPosition()
        axis = self.data.axes[self.combobox.currentIndex()]
        value = axis.axis[index]

        text = '%.2f  %s' % (value, axis.units)
        self.value_label.setText(text)

    def _update_slider_silently(self, index):

        self.slider.blockSignals(True)

        axis = self.data.axes[self.combobox.currentIndex()]
        maximum = len(axis.axis) - 1
        self.slider.setMaximum(maximum)
        self.slider.setSliderPosition(index)

        self.slider.blockSignals(False)

    def _update_spinbox_silently(self, index):

        self.spinbox.blockSignals(True)

        axis = self.data.axes[self.combobox.currentIndex()]
        maximum = len(axis.axis) - 1
        self.spinbox.setMaximum(maximum)
        self.spinbox.setValue(index)

        self.spinbox.blockSignals(False)

    def _setup_combobox(self):

        for index in range(3):
            axis_label = self.data.axes[index].label
            self.combobox.addItem(str(axis_label))

    def _setup(self):

        self._setup_combobox()
        self._update_slice_label()
        self.change_axis(0)

    def _connect(self):

        self.slider.valueChanged.connect(self.change_slice)
        self.spinbox.valueChanged.connect(self.change_slice)
        self.combobox.currentIndexChanged.connect(self.change_axis)
        self.transpose_button.clicked.connect(self.trigger_transpose)



# Python Imports
from os.path import abspath

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtGui import QTextDocument
from PyQt5.QtCore import QDir

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab
from kmap.config.config import config


class FileTab(Tab):

    def __init__(self, path, title, richText=False):

        super(FileTab, self).__init__()

        self.title = title
        self.path = path
        self.richText = richText

    def reload_text(self):

        with open(self.path, 'r') as file:
            text = file.read()

        self.refresh_display(text)

    def refresh_display(self, text):

        if self.richText:
            self.display.setHtml(text)
        else:
            self.display.setPlainText(text)

    def find_next(self):

        self.display.find(self.line_edit.text())

    def find_prev(self):

        self.display.find(self.line_edit.text(), QTextDocument.FindBackward)

    def get_title(self):

        return self.title

    def set_title(self, title):

        self.title = title

    def _setup(self):

        label = self.title_line_edit.text()
        self.title_line_edit.setText(label + self.path)

    def _connect(self):

        self.find_next_button.clicked.connect(self.find_next)
        self.find_prev_button.clicked.connect(self.find_prev)
        self.reload_button.clicked.connect(self.reload_text)


# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/fileviewertab.ui')
FileViewerTab_UI, _ = uic.loadUiType(UI_file)


class FileViewerTab(FileTab, FileViewerTab_UI):

    def __init__(self, path, title, richText=False):

        # Setup GUI
        super(FileViewerTab, self).__init__(path, title, richText=richText)
        self.setupUi(self)
        self._connect()

        self._setup()
        self.reload_text()


# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/fileeditortab.ui')
FileEditorTab_UI, _ = uic.loadUiType(UI_file)


class FileEditorTab(FileTab, FileEditorTab_UI):

    def __init__(self, path, title):

        # Setup GUI
        super(FileEditorTab, self).__init__(path, title, richText=False)
        self.setupUi(self)
        self._connect()

        self._setup()
        self.reload_text()

    def save(self):

        if self.richText:
            text = self.display.toHtml()

        else:
            text = self.display.toPlainText()

        with open(self.path, 'wt') as file:
            file.write(text)

    def _connect(self):

        FileTab._connect(self)

        self.save_button.clicked.connect(self.save)



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import QDir

# Own Imports
from kmap import __directory__
from kmap.model.colormap_model import ColormapModel
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/colormap.ui')
Colormap_UI, _ = uic.loadUiType(UI_file)


class Colormap(QWidget, Colormap_UI):

    def __init__(self, plot_item):

        # Setup GUI
        super(Colormap, self).__init__()
        self.setupUi(self)
        self._connect()

        # Path for the .json file containing the colormaps
        self.path = __directory__ + config.get_key('paths', 'colormap')
        self.model = ColormapModel(plot_item)

        self.load_colormaps()

    def load_colormaps(self):

        try:
            self.model.load_colormaps(self.path)

        except Exception:

            log = logging.getLogger('kmap')

            log.error('Colormaps could not be loaded')
            log.error(traceback.format_exc())

        self._update_combobox()
        self.set_default_colormap()

    def add_colormap(self):

        name = self.line_edit.text()

        if name:
            self.model.add_colormap_from_plot(name)
            self._update_combobox()
            self.set_colormap(name)

    def remove_colormap(self):

        name = self.combobox.currentText()

        if name:
            self.model.remove_colormap(name)
            self._update_combobox()
            self.set_default_colormap()

    def save(self):

        try:
            self.model.save_colormaps(self.path)

        except Exception:

            log = logging.getLogger('kmap')

            log.error('Colormaps could not be saved')
            log.error(traceback.format_exc())

            os.remove(path_temp)

    def set_colormap(self, name):

        self.model.set_current_colormap(name)
        self.combobox.setCurrentText(name)

    def set_default_colormap(self):

        default_colormap = config.get_key('colormap', 'default')
        self.set_colormap(default_colormap)

    def _update_combobox(self):

        self.combobox.blockSignals(True)

        self.combobox.clear()
        for colormap in self.model.colormaps:
            self.combobox.addItem(colormap.name)

        self.combobox.blockSignals(False)

    def _connect(self):

        self.combobox.currentTextChanged.connect(self.set_colormap)
        self.remove_button.clicked.connect(self.remove_colormap)
        self.add_button.clicked.connect(self.add_colormap)
        self.save_button.clicked.connect(self.save)
        self.reload_button.clicked.connect(self.load_colormaps)



from kmap import __version__, __date__


class MainWindowModel():

    def __init__(self, controller):

        self.controller = controller

    def get_about_text(self, path):

        title = 'kMap.py'
        with open(path, 'r') as file:
            text = file.read() % (__version__, __date__)

        return title, text




# Python Imports
import numpy as np
import pyqtgraph as pg

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import QDir

# Own Imports
from kmap import __directory__
from kmap.config.config import config
from kmap.library.misc import normalize
from kmap.model.crosshair_model import CrosshairAnnulusModel
from kmap.controller.crosshairroi import CrosshairROIBase

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/crosshairannulus.ui')
CrosshairAnnulus_UI, _ = uic.loadUiType(UI_file)


class CrosshairAnnulusBase(CrosshairROIBase):

    def __init__(self, plot_item):

        super().__init__(plot_item)

    def enable_annulus(self, enable):

        self.annulus.setVisible(enable)

    def dragging_annulus(self):

        self.dragging_annulus = True

    def resize_annulus_from_drag(self):

        # There is no signal for finished dragging, but starting it
        if not self.dragging_annulus:
            return

        self.dragging_annulus = False

        width = (self.annulus.size()[0] - self.roi.size()[0]) / 2

        if width <= 0:
            width = 0.01

        self.model.set_width(width)

        self.update()

    def resize_annulus_from_spinbox(self):

        width = self.width_spinbox.value()

        if width <= 0:
            width = 0.01

        self.model.set_width(width)

        self.update()

    def move_crosshair_from_spinbox(self):

        update = super().move_crosshair_from_spinbox()

        if update:
            x, y = self.model.x, self.model.y
            self.annulus.setPos(x, y)

            return True

        else:
            return False

    def change_color(self, index):

        CrosshairROIBase.change_color(self, index)

        self.annulus.setPen(self.colors[index])

    def update(self):

        x, y = self.model.x, self.model.y
        radius, width = self.model.radius, self.model.width
        large_radius = radius + width
        # Update Spinboxes silently
        self.roi_spinbox.blockSignals(True)
        self.width_spinbox.setValue(width)
        self.roi_spinbox.blockSignals(False)

        # Update Crosshair Position and Size
        self.annulus.setPos([x - large_radius, y - large_radius])
        self.annulus.setSize([2 * large_radius, 2 * large_radius])

        super().update()

    def update_label(self):

        plot_data = self.plot_item.get_plot_data()

        super().update_label()

        if plot_data == None:
            intensity = 0

        else:
            cut = self.model.cut_from_data(plot_data, region='ring')

            # Normalize by dividing by the number of non nan elements
            if config.get_key('crosshair', 'normalized_intensity') == 'True':
                intensity = normalize(cut.data)

            else:
                intensity = np.nansum(cut.data)

        if abs(intensity) > 1000:
            self.ring_value_label.setText('%.2fk' % (intensity / 1000))

        else:
            self.ring_value_label.setText('%.2f' % intensity)

    def _set_model(self, model=None):

        if model is None:
            self.model = CrosshairAnnulusModel(x=0, y=0, radius=0.2, width=0.1)

        else:
            self.model = model

    def _setup(self):

        CrosshairROIBase._setup(self)

        x, y = self.model.x, self.model.y
        radius, width = self.model.radius, self.model.width
        large_radius = radius + width
        self.annulus = pg.CircleROI([x - large_radius, y - large_radius],
                                    size=[2 * large_radius, 2 * large_radius],
                                    movable=False,
                                    rotatable=False,
                                    resizable=True,
                                    removable=False,
                                    pen='k',
                                    handlePen='r')
        self.plot_item.addItem(self.annulus)

    def _connect(self):

        CrosshairROIBase._connect(self)

        self.enable_annulus_checkbox.stateChanged.connect(self.enable_annulus)

        self.width_spinbox.valueChanged.connect(
            self.resize_annulus_from_spinbox)
        self.annulus.sigRegionChangeFinished.connect(
            self.resize_annulus_from_drag)
        self.annulus.sigRegionChangeStarted.connect(self.dragging_annulus)


class CrosshairAnnulus(CrosshairAnnulusBase, CrosshairAnnulus_UI):

    def __init__(self, plot_item):

        # Setup GUI
        super(CrosshairAnnulus, self).__init__(plot_item)
        self.setupUi(self)
        self._setup()
        self._connect()

        self.dragging_roi = False
        self.enable_roi(False)
        self.dragging_annulus = False
        self.enable_annulus(False)

        self.enable(False)
        self.update_label()



import unittest
from PyQt5.QtWidgets import QApplication
from kmap import kMap


class TestMainWindow(unittest.TestCase):

    app = kMap.kMap([])

    def test_default_state(self):

        self.app.test()



import unittest
from kmap.library.orbitaldata import OrbitalData
from kmap import __directory__


class TestOrbitalData(unittest.TestCase):

    def test_default_initialization(self):

        name = 'Test'
        ID = 1
        meta_data = {'origin': 'selfmade'}
        with open(__directory__ +
                  '/../example/data/5A_MO_73.cube', 'r') as f:
            file = f.read()

        data = OrbitalData(file, ID, name=name,
                           meta_data=meta_data)

        self.assertEqual(data.name, name)
        self.assertEqual(data.ID, ID)
        self.assertEqual(data.meta_data, meta_data)

    def test_file_initialization(self):

        ID = 1
        data = OrbitalData.init_from_file(__directory__ +
                                          '/../example/data/' +
                                          '5A_MO_73.cube', ID)

        self.assertEqual(data.name, '5A_MO_73')
        self.assertEqual(data.ID, ID)
        self.assertEqual(data.meta_data, {
                         'origin': 'Cube file generated by NWChem',
                         'alias': 'pentacene_MO_73'})

    def test_online_initialization(self):

        ID = 1
        data = OrbitalData.init_from_online('http://143.50.77.12/' +
                                            'OrganicMolecule/B3LYP/5A/' +
                                            'charge0mult1/5A_MO_73', ID)

        self.assertEqual(data.name, '5A_MO_73')
        self.assertEqual(data.ID, ID)
        self.assertEqual(data.meta_data, {
                         'origin': 'Cube file generated by NWChem',
                         'alias': 'pentacene_MO_73'})



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal, QDir
from PyQt5.QtWidgets import QWidget

# Third Party Imports
import numpy as np
import pyqtgraph as pg

# Own Imports
from kmap import __directory__
from kmap.config.config import config
from kmap.library.misc import normalize
from kmap.model.crosshair_model import CrosshairModel

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/crosshair.ui')
Crosshair_UI, _ = uic.loadUiType(UI_file)


class CrosshairBase(QWidget):

    crosshair_changed = pyqtSignal()

    def __init__(self, plot_item):

        super(CrosshairBase, self).__init__()

        self.plot_item = plot_item
        self._set_model()

        self.colors = ['k', 'w', 'r', 'b', 'g', 'c', 'm', 'y']

    def enable(self, enable):

        self.v_line.setVisible(enable)
        self.h_line.setVisible(enable)

    def update_label(self):

        plot_data = self.plot_item.get_plot_data()

        if plot_data == None:
            intensity = 0

        elif np.isnan(plot_data.data).all():
            intensity = np.nan

        else:
            cut = self.model.cut_from_data(plot_data, region='center')

            # Normalize by dividing by the number of non nan elements
            if config.get_key('crosshair', 'normalized_intensity') == 'True':
                intensity = normalize(cut.data)

            else:
                intensity = np.nansum(cut.data)

        if abs(intensity) > 1000:
            self.point_value_label.setText('%.2fk' % (intensity / 1000))

        else:
            self.point_value_label.setText('%.2f' % intensity)

        x = self.model.x
        y = self.model.y

        self.distance_value_label.setText('%.2f' % np.sqrt(x**2 + y**2))

    def move_crosshair_from_drag(self):

        x, y = self.v_line.value(), self.h_line.value()
        self.model.set_position(x, y)

        self.update()

    def move_crosshair_from_click(self, event):

        if not self.enable_crosshair_checkbox.isChecked():
            return False

        click_pos = self.plot_item.view.vb.mapSceneToView(
            event.scenePos())
        self.model.set_position(click_pos.x(), click_pos.y())

        self.update()

        return True

    def move_crosshair_from_spinbox(self):

        x, y = self.x_spinbox.value(), self.y_spinbox.value()
        self.model.set_position(x, y)

        self.update()

    def change_color(self, index):

        self.h_line.setPen(self.colors[index])
        self.v_line.setPen(self.colors[index])

    def update(self):

        x, y = self.model.x, self.model.y

        # Update Spinboxes silently
        self.x_spinbox.blockSignals(True)
        self.y_spinbox.blockSignals(True)
        self.x_spinbox.setValue(x)
        self.y_spinbox.setValue(y)
        self.x_spinbox.blockSignals(False)
        self.y_spinbox.blockSignals(False)

        # Update Crosshair Position
        self.v_line.setValue(x)
        self.h_line.setValue(y)

        # Emit Changed Signal
        self.crosshair_changed.emit()

    def _set_model(self, model=None):

        if model is None:
            self.model = CrosshairModel(x=0, y=0)

        else:
            self.model = model

    def _setup(self):

        self.v_line = pg.InfiniteLine(movable=True, angle=90, pen='k',
                                      bounds=[self.x_spinbox.minimum(),
                                              self.x_spinbox.maximum()])
        self.h_line = pg.InfiniteLine(movable=True, angle=0, pen='k',
                                      bounds=[self.y_spinbox.minimum(),
                                              self.y_spinbox.maximum()])
        self.v_line.setValue(self.model.x)
        self.h_line.setValue(self.model.y)
        self.plot_item.addItem(self.v_line)
        self.plot_item.addItem(self.h_line)

    def _connect(self):

        self.x_spinbox.valueChanged.connect(self.move_crosshair_from_spinbox)
        self.y_spinbox.valueChanged.connect(self.move_crosshair_from_spinbox)

        self.enable_crosshair_checkbox.stateChanged.connect(self.enable)

        self.v_line.sigDragged.connect(self.move_crosshair_from_drag)
        self.h_line.sigDragged.connect(self.move_crosshair_from_drag)
        self.plot_item.view.scene().sigMouseClicked.connect(
            self.move_crosshair_from_click)

        self.color_combobox.currentIndexChanged.connect(self.change_color)


class Crosshair(CrosshairBase, Crosshair_UI):

    def __init__(self, plot_item):

        # Setup GUI
        super(Crosshair, self).__init__(plot_item)
        self.setupUi(self)

        self._setup()
        self._connect()

        self.enable(False)
        self.update_label()



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab

# Load .ui File
UI_file = __directory__ + '/ui/lmfitplottab.ui'
LMFitPlotTab_UI, _ = uic.loadUiType(UI_file)


class LMFitPlotTab(Tab, LMFitPlotTab_UI):

    def __init__(self, results, orbitals, axis, *args, **kwargs):

        self.results = results
        self.orbitals = orbitals
        self.x_axis = axis

        # Setup GUI
        super(LMFitPlotTab, self).__init__(*args, **kwargs)
        self.setupUi(self)

        self._connect()

        self.refresh_plot()

    def export_to_txt(self):

        data = self.plot_item.get_data()

        text = ''

        for data_set in data:
            name = data_set['name']
            x = data_set['x'] 
            y = data_set['y']
            text += '# '+name+'\n'
            for xi, yi in zip(x,y):
                text += '%g  %g \n'%(xi, yi)
            text += '\n'              

        return text
        
    def refresh_plot(self):

        possible_params = ['w_', 'phi_', 'theta_', 'psi_']
        possible_labels = ['Weight [1]', 'Phi [°]', 'Theta [°]', 'Psi [°]']
        param = possible_params[self.parameter_combobox.currentIndex()]

        self.plot_item.clear()

        for orbital in self.orbitals:
            y = [result.params[param + str(orbital.ID)]
                 for result in self.results]
            x = self.x_axis.axis
            title = orbital.name

            self.plot_item.plot(x, y, title)

        self.plot_item.set_label('%s [%s]' % (self.x_axis.label,
                                              self.x_axis.units),
                                 possible_labels[self.parameter_combobox.currentIndex()])

    def _connect(self):

        self.refresh_button.clicked.connect(self.refresh_plot)



# Python Imports
import os
import unittest

# Third Party Imports
import numpy as np
import numpy.testing as npt

# Own Imports
from kmap.model.lmfit_model import LMFitModel
from kmap.model.crosshair_model import CrosshairAnnulusModel
from kmap.library.sliceddata import SlicedData
from kmap.library.orbitaldata import OrbitalData
from kmap.library.misc import step_size_to_num
from kmap.config.config import config


class TestLMFitModel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        file_path = os.path.dirname(os.path.realpath(
            __file__))
        input_path = file_path + '/../../example/data/'

        sliced_path = input_path + 'example5_6584.hdf5'
        cls.sliced_data = SlicedData.init_from_hdf5(sliced_path)

        orbital_paths = ['PTCDA_C.cube', 'PTCDA_D.cube',
                         'PTCDA_E.cube', 'PTCDA_F.cube']
        cls.orbitals = [OrbitalData.init_from_file(
            input_path + path, ID) for ID, path in enumerate(orbital_paths)]

        cls.expected = np.loadtxt(file_path + '/output/weights_PTCDA')
        cls.background_expected = np.loadtxt(
            file_path + '/output/background_expected')

    def setUp(self):

        self.lmfit = LMFitModel(
            TestLMFitModel.sliced_data, TestLMFitModel.orbitals)
        self.crosshair = CrosshairAnnulusModel()

    def test_set_crosshair(self):

        self.lmfit.set_crosshair(self.crosshair)

        self.assertEqual(self.lmfit.crosshair, self.crosshair)

    def test_set_axis(self):

        step_size = 0.24
        range_ = [-3, 3]
        axis = np.linspace(
            *range_, num=step_size_to_num(range_, step_size), endpoint=True)

        self.lmfit.set_axis_by_step_size(range_, step_size)
        npt.assert_almost_equal(self.lmfit.get_sliced_kmap(0).x_axis, axis)

    def test_set_slices(self):

        self.lmfit.set_slices([1, 2, 3])
        self.lmfit.set_slices([0, 1, 2, 3], combined=True)

    def test_set_background_equation(self):

        self.lmfit.set_background_equation('np.exp(a)')
        npt.assert_equal(self.lmfit.background_equation, ['np.exp(a)', ['a']])

        self.assertRaises(
            ValueError, self.lmfit.set_background_equation, 'np.exp(a')

    def test_parameters(self):

        self.lmfit.set_background_equation('np.exp(a)')

        self.assertEqual(self.lmfit.parameters['w_1'].min, 0)
        self.assertEqual(self.lmfit.parameters['a'].value, 0)
        self.assertEqual(self.lmfit.parameters['E_kin'].max, 150)

    def test_edit_parameter(self):

        self.lmfit.edit_parameter('w_1', value=1.5)
        self.assertEqual(self.lmfit.parameters['w_1'].value, 1.5)

    def test_background(self):

        range_, dk = [-3.0, 3.0], 0.025
        self.lmfit.set_axis_by_step_size(range_, dk)

        self.lmfit.set_background_equation(
            '(np.exp(-x**2-y**2)-np.exp(-(x-1)**2-(y-1)**2))/2')

        background = self.lmfit._get_background(variables={})

        npt.assert_almost_equal(background, TestLMFitModel.background_expected)

    def test_settings(self):

        lmfit_new = LMFitModel(
            TestLMFitModel.sliced_data, TestLMFitModel.orbitals)

        lmfit_new.set_settings(self.lmfit.get_settings())

    def test_PTCDA(self):

        if float(config.get_key('orbital', 'dk3D')) != 0.12:
            print(
                'WARNING: Test \'test_PTCDA\' from the ' +
                '\'test_lmfit\' module has not been run. It requires ' +
                '\'dk3D\' setting from the \'cube\' category to be ' +
                'to 0.12.')
            return

        # Set certain parameters not being fitted but desired to be changed
        range_, dk = [-3.0, 3.0], 0.04
        self.lmfit.set_axis_by_step_size(range_, dk)
        self.lmfit.set_polarization('toroid', 'p')
        self.lmfit.set_background_equation('c')

        # Set certain fit parameter to desired value
        self.lmfit.edit_parameter('E_kin', value=27.2)
        self.lmfit.edit_parameter('alpha', value=40)
        self.lmfit.edit_parameter('c', value=1, vary=True)

        # Activate fitting for all weights (i is a dummy ID used in setUpClass)
        for i in [0, 1, 2, 3]:
            self.lmfit.edit_parameter('w_' + str(i), vary=True)

        # Set slices to be used
        self.lmfit.set_slices('all', combined=False)

        results = self.lmfit.fit()

        # Test results
        weights = np.array([[result[1].params['w_0'].value,
                             result[1].params['w_1'].value,
                             result[1].params['w_2'].value,
                             result[1].params['w_3'].value]
                            for result in results]).T

        npt.assert_almost_equal(weights, TestLMFitModel.expected, decimal=5)



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import Qt, pyqtSignal, QDir
from PyQt5.QtWidgets import QWidget, QHeaderView, QHBoxLayout

# Own Imports
from kmap import __directory__
from kmap.controller.orbitaltablerow import OrbitalTableRow

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/orbitaltable.ui')
OrbitalTable_UI, _ = uic.loadUiType(UI_file)


class OrbitalTable(QWidget, OrbitalTable_UI):

    orbital_changed = pyqtSignal(int)
    orbital_removed = pyqtSignal(int)
    orbital_selected = pyqtSignal(int)

    def __init__(self, *args, **kwargs):

        # Setup GUI
        super(OrbitalTable, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()

        self.rows = []

    def add_orbital(self, orbital, orientation):

        new_row = OrbitalTableRow(orbital, orientation)

        self._add_row(new_row)

    def remove_orbital_by_ID(self, ID):

        index = self._ID_to_row_index(ID)

        self.table.removeRow(index)
        del self.rows[index]

        self.orbital_removed.emit(ID)

    def get_parameters_by_ID(self, ID):

        index = self._ID_to_row_index(ID)

        return self.rows[index].get_parameters()

    def get_use_by_ID(self, ID):

        index = self._ID_to_row_index(ID)

        return self.rows[index].get_use()

    def change_parameter(self, ID, parameter, value):

        self._update_selected_rows(parameter, value)

        self.orbital_changed.emit(ID)

    def change_use(self, ID):

        self.orbital_changed.emit(ID)

    def _add_row(self, new_row):

        row_index = len(self.rows)

        self.rows.append(new_row)

        self.table.insertRow(row_index)
        self._embed_row(new_row, row_index)
        self._connet_row(new_row)
        # "Select" newly added row to trigger miniplots
        self.table.cellClicked.emit(row_index, 2)

    def _ID_to_row_index(self, ID):

        for index, row in enumerate(self.rows):
            if row.ID == ID:
                return index

        raise IndexError('No data with this ID')

    def _update_selected_rows(self, parameter, value):

        selected_rows = self.table.selectionModel().selectedIndexes()
        for selected_row in selected_rows:
            row = self.rows[selected_row.row()]
            row.update_parameter_silently(parameter, value)

    def _setup(self):

        widths = [40, 40, 0, 90, 80, 80, 80, 45]

        for col, width in enumerate(widths):
            self.table.setColumnWidth(col, width)

        self.table.horizontalHeader().setResizeMode(2, QHeaderView.Stretch)

    def _embed_row(self, row, row_index):

        # To center checkbox, embed it inside a layout inside a widget
        layout = QHBoxLayout()
        layout.addWidget(row.use)
        layout.setAlignment(Qt.AlignCenter)
        layout.setContentsMargins(0, 0, 0, 0)
        center_widget = QWidget()
        center_widget.setLayout(layout)

        widgets = [row.button, row.ID_label, row.name_label, row.weight,
                   row.phi, row.theta, row.psi, center_widget]

        for col, widget in enumerate(widgets):
            self.table.setCellWidget(row_index, col, widget)

    def _row_selected(self, row):

        self.orbital_selected.emit(self.rows[row].ID)

    def _connet_row(self, row):

        row.row_removed.connect(self.remove_orbital_by_ID)
        row.parameter_changed.connect(self.change_parameter)
        row.use_changed.connect(self.change_use)

        self.table.cellClicked.connect(self._row_selected)



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import QDir

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab
from kmap.model.sliceddatatab_model import SlicedDataTabModel
from kmap.controller.matplotlibwindow import MatplotlibWindow
from kmap.controller.interpolation import Interpolation
from kmap.controller.dataslider import DataSlider
from kmap.controller.crosshairannulus import CrosshairAnnulus
from kmap.controller.pyqtgraphplot import PyQtGraphPlot
from kmap.controller.colormap import Colormap


# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/sliceddatatab.ui')
SlicedDataTab_UI, _ = uic.loadUiType(UI_file)


class SlicedDataTab(Tab, SlicedDataTab_UI):

    def __init__(self, model):

        self.model = model

        # Setup GUI
        super(SlicedDataTab, self).__init__()
        self.setupUi(self)
        self._setup()
        self._connect()

        self.change_axis(0)
        self.change_slice(0)

    @classmethod
    def init_from_URLs(cls, URLs):

        model = SlicedDataTabModel()
        model.load_data_from_URLs(URLs)

        return cls(model)

    @classmethod
    def init_from_URL(cls, URL):

        model = SlicedDataTabModel()
        model.load_data_from_URL(URL)

        return cls(model)

    @classmethod
    def init_from_cube(cls, URL):

        model = SlicedDataTabModel()
        model.load_data_from_cube(URL)

        return cls(model)
        
    @classmethod
    def init_from_path(cls, path):

        model = SlicedDataTabModel()
        model.load_data_from_path(path)

        return cls(model)

    def get_data(self):

        return self.model.data

    def get_axis(self):

        return self.slider.get_axis()

    def change_slice(self, index=-1):

        axis = self.slider.get_axis()
        slice_index = index if index != -1 else self.slider.get_index()
        data = self.model.change_slice(slice_index, axis)

        data = self.interpolation.interpolate(data)
        data = self.interpolation.smooth(data)

        self.plot_item.plot(data)
        self.crosshair.update_label()

    def change_axis(self, axis):

        # 'axes' is a copy of all axes except the one with index 'axis'
        axes = [a for i, a in enumerate(self.model.data.axes) if i != axis]

        index = self.slider.get_index()
        data = self.model.change_slice(index, axis)

        self.plot_item.set_labels(axes[1], axes[0])
        self.interpolation.set_label(axes[1], axes[0])
        self.plot_item.plot(data)

    def crosshair_changed(self):

        self.crosshair.update_label()

    def display_in_matplotlib(self):

        data = self.model.displayed_plot_data
        LUT = self.plot_item.get_LUT()

        window = MatplotlibWindow(data, LUT=LUT)

        return window

    def closeEvent(self, event):

        del self.model

        Tab.closeEvent(self, event)

    def to_string(self):

        text = self.model.to_string()

        return text

    def get_displayed_plot_data(self):

        return self.model.displayed_plot_data

    def get_crosshair(self):

        return self.crosshair

    def get_plot_labels(self):

        bottom = self.plot_item.get_label('bottom')
        left = self.plot_item.get_label('left')
        return bottom, left

    def transpose(self, constant_axis):

        self.model.transpose(constant_axis)
        self.change_axis(self.slider.get_axis())

    def _setup(self):

        self.slider = DataSlider(self.model.data)
        self.crosshair = CrosshairAnnulus(self.plot_item)
        self.colormap = Colormap(self.plot_item)
        self.interpolation = Interpolation()

        layout = self.scroll_area.widget().layout()
        layout.insertWidget(0, self.slider)
        layout.insertWidget(1, self.colormap)
        layout.insertWidget(2, self.crosshair)
        layout.insertWidget(3, self.interpolation)

        # Set Title
        data = self.model.data

        if data:
            id_ = data.ID

            if 'alias' in data.meta_data:
                text = data.meta_data['alias']
            else:
                text = data.name

            self.title = '%s (%i)' % (text, id_)

        else:
            self.title = 'NO DATA'

    def _connect(self):

        self.crosshair.crosshair_changed.connect(self.crosshair_changed)
        self.slider.slice_changed.connect(self.change_slice)
        self.slider.axis_changed.connect(self.change_axis)
        self.slider.tranpose_triggered.connect(self.transpose)
        self.interpolation.smoothing_changed.connect(self.change_slice)
        self.interpolation.interpolation_changed.connect(self.change_slice)



"""Unittests for the PlotData class.
"""

# Python Imports
import unittest

# Third Party Imports
import numpy as np
import numpy.testing as npt

# Own Imports
from kmap.library.plotdata import PlotData


class TestPlotData(unittest.TestCase):

    def setUp(self):

        self.data = [[1, 2, 3], [4, 5, 6]]
        self.range_ = [[1, 2], [-1, 1]]
        self.plot_data = PlotData(self.data, self.range_)

    def test_correct_initialization(self):

        npt.assert_equal(self.plot_data.data, self.data)
        npt.assert_equal(self.plot_data.range, self.range_)
        npt.assert_equal(self.plot_data.data.shape, (2, 3))
        npt.assert_equal(self.plot_data.x_axis, [1., 1.5, 2.])
        npt.assert_equal(self.plot_data.y_axis, [-1., 1.])
        npt.assert_equal(self.plot_data.step_size, [0.5, 2.])

        self.data = [[1, 2, np.nan], [4, np.inf, 6]]
        plot_data = PlotData(self.data, self.range_)

        npt.assert_equal(
            plot_data.data, [[1, 2, np.nan], [4, np.nan, 6]])

    def test_incorrect_data_shape(self):

        self.data = [1, 2, 3, 4, 5, 6]
        self.assertRaises(TypeError, PlotData, self.data, self.range_)

    def test_too_small_data(self):

        self.data = [[1], [2]]
        self.assertRaises(TypeError, PlotData, self.data, self.range_)

    def test_incorrect_range_shape(self):

        self.range_ = [[1, 2]]
        self.assertRaises(TypeError, PlotData, self.data, self.range_)

    def test_incorrect_range_values(self):

        self.range_ = [[1, 2], [np.nan, 1]]
        self.assertRaises(ValueError, PlotData, self.data, self.range_)

        self.range_ = [[1, 2], [np.inf, 1]]
        self.assertRaises(ValueError, PlotData, self.data, self.range_)

    def test_basic_interpolation(self):

        new_x_axis = [1.5, 2]
        new_y_axis = [-0.5, 0, 0.5]
        new_data = self.plot_data.interpolate(new_x_axis, new_y_axis)
        npt.assert_equal(
            new_data.data, [[2.75, 3.75], [3.5, 4.5], [4.25, 5.25]])
        npt.assert_equal(self.plot_data.data, self.data)

        new_x_axis = [-2, 1]
        new_y_axis = [0, -0.5, -10]
        new_data = self.plot_data.interpolate(new_x_axis, new_y_axis)
        npt.assert_equal(
            new_data.data, [[np.nan, 2.5], [np.nan, 1.75], [np.nan, np.nan]])

        new_x_axis = [1, 1.25, 1.5, 2]
        new_y_axis = [-1, -0.25, 0.5, 1]
        new_data = self.plot_data.interpolate(new_x_axis, new_y_axis)
        npt.assert_equal(
            new_data.data, [[1., 1.5, 2., 3.],
                            [2.125, 2.625, 3.125, 4.125],
                            [3.25, 3.75, 4.25, 5.25],
                            [4., 4.5, 5., 6.]])

    def test_update_interpolation(self):

        new_x_axis = [1.5, 2]
        new_y_axis = [-0.5, 0, 0.5]
        new_data = self.plot_data.interpolate(
            new_x_axis, new_y_axis, update=True)
        npt.assert_equal(
            new_data.data, [[2.75, 3.75], [3.5, 4.5], [4.25, 5.25]])
        npt.assert_equal(self.plot_data.data, [
                         [2.75, 3.75], [3.5, 4.5], [4.25, 5.25]])

    def test_increase_range(self):

        new_x_axis = [0, 1, 2, 3]
        new_y_axis = [-1, 0, 1, 2]
        new_data = self.plot_data.interpolate(
            new_x_axis, new_y_axis, update=True)
        npt.assert_equal(new_data.data, [[np.nan, 1., 3., np.nan],
                                         [np.nan, 2.5, 4.5, np.nan],
                                         [np.nan, 4., 6., np.nan],
                                         [np.nan, np.nan, np.nan, np.nan]])
        npt.assert_equal(self.plot_data.x_axis, [0, 1, 2, 3])
        npt.assert_equal(self.plot_data.y_axis, [-1, 0, 1, 2])

    def test_add(self):

        data = [[7, 8, 9], [10, 11, 12]]
        range_ = [[1, 2], [-1, 1]]

        second_plot_data = PlotData(data, range_)

        added_plot_data = self.plot_data + second_plot_data
        npt.assert_equal(added_plot_data.data, [[8, 10, 12], [14, 16, 18]])

        added_plot_data = self.plot_data + 4
        npt.assert_equal(added_plot_data.data, [[5, 6, 7], [8, 9, 10]])

    def test_incorrect_add(self):

        data = [[7, 8, 9], [10, 11, 12]]
        range_ = [[1, 3], [-1, 1]]

        second_plot_data = PlotData(data, range_)

        self.assertRaises(ValueError, self.plot_data.__add__, second_plot_data)
        self.assertRaises(TypeError, self.plot_data.__add__, 'string')

    def test_sub(self):

        data = [[7, 8, 9], [10, 11, 12]]
        range_ = [[1, 2], [-1, 1]]

        second_plot_data = PlotData(data, range_)

        subed_plot_data = self.plot_data - second_plot_data
        npt.assert_equal(subed_plot_data.data, [[-6, -6, -6], [-6, -6, -6]])

        subed_plot_data = self.plot_data - np.array(data)
        npt.assert_equal(subed_plot_data.data, [[-6, -6, -6], [-6, -6, -6]])

        subed_plot_data = self.plot_data - 4
        npt.assert_equal(subed_plot_data.data, [[-3, -2, -1], [0, 1, 2]])

    def test_incorrect_sub(self):

        data = [[7, 8, 9], [10, 11, 12]]
        range_ = [[1, 3], [-1, 1]]

        second_plot_data = PlotData(data, range_)

        self.assertRaises(ValueError, self.plot_data.__sub__, second_plot_data)
        self.assertRaises(TypeError, self.plot_data.__sub__, 'string')

    def test_mul(self):

        data = [[2, 2, 2], [3, 3, 3]]
        range_ = [[1, 2], [-1, 1]]

        second_plot_data = PlotData(data, range_)

        mul_plot_data = self.plot_data * second_plot_data
        npt.assert_equal(mul_plot_data.data, [[2, 4, 6], [12, 15, 18]])

        mul_plot_data = second_plot_data * self.plot_data
        npt.assert_equal(mul_plot_data.data, [[2, 4, 6], [12, 15, 18]])

        mul_plot_data = self.plot_data * 4
        npt.assert_equal(mul_plot_data.data, [[4, 8, 12], [16, 20, 24]])

    def test_incorrect_mul(self):

        data = [[7, 8, 9], [10, 11, 12]]
        range_ = [[1, 3], [-1, 1]]

        second_plot_data = PlotData(data, range_)

        self.assertRaises(ValueError, self.plot_data.__mul__, second_plot_data)
        self.assertRaises(TypeError, self.plot_data.__mul__, 'string')
        self.assertRaises(
            ValueError, self.plot_data.__rmul__, second_plot_data)
        self.assertRaises(TypeError, self.plot_data.__rmul__, 'string')

    def test_pow(self):

        pow_plot_data = self.plot_data ** 2
        npt.assert_equal(pow_plot_data.data, [[1, 4, 9], [16, 25, 36]])


if __name__ == '__main__':
    unittest.main()



class Database():

    def __init__(self, path):

        self.molecules = []
        self.URL = ''

        self._load_molecules(path)

    def get_molecule_by_ID(self, ID):

        for molecule in self.molecules:
            if molecule.ID == ID:
                return molecule

        return None

    def _load_molecules(self, path):

        with open(path, 'r') as file:

            first_line = file.readline()
            self.URL = first_line.split('=')[1][:-1]

            lines = []
            lines.append(file.readline())

            # Loop over all Molecules
            for line in file:
                if line[:11] == 'moleculeID=':
                    new_molecule = Molecule(lines, URL=self.URL)
                    self.molecules.append(new_molecule)

                    lines = []

                lines.append(line)

            # Append last molecule
            new_molecule = Molecule(lines, URL=self.URL)
            self.molecules.append(new_molecule)

class Molecule():

    def __init__(self, lines, URL=''):

        self.ID = int(lines[0].split('=')[1])
        self.URL = URL + lines[2].split('=')[2][:-1]

        parts = lines[1].split(',')

        self.short_name = parts[0].split('=')[1]
        self.full_name = parts[1].split('=')[1]
        self.formula = parts[2].split('=')[1]
        self.orientation = parts[3].split('=')[1]
        self.charge = int(parts[4].split('=')[1])
        self.magnetic_moment = float(parts[5].split('=')[1])
        self.IP = float(parts[6].split('=')[1])
        self.EA = float(parts[7].split('=')[1])
        self.basis_set = parts[8].split('=')[1]
        self.XC_functional = parts[9].split('=')[1][:-1]

        self.orbitals = []
        for index, line in enumerate(lines[3:], 1):
            new_orbital = Orbital(index,self.ID, line, self.URL, self.orientation, self.short_name)
            self.orbitals.append(new_orbital)

    def to_string(self):

        ID = 'ID: %i\n' % self.ID
        URL = 'URL: %s\n' % self.URL
        small_name = 'Short Name: % s\n' % self.short_name
        full_name = 'Full Name: %s\n' % self.full_name
        formula = 'Formula: %s\n' % self.formula
        orientation = 'Orientation: %s\n' % self.orientation
        charge = 'Charge: %i\n' % self.charge
        magnetic_moment = 'Magnetic Moment: %.3f\n' % self.magnetic_moment
        IP = 'IP: %.2f\n' % self.IP
        EA = 'EA: %.2f\n' % self.EA
        basis_set = 'Basis Set: %s\n' % self.basis_set
        XC_functional = 'XC-Functional: %s\n' % self.XC_functional
        length = 'Number of Orbitals: %i' % len(self.orbitals)

        return (ID + URL + small_name + full_name + formula +
                orientation + charge + magnetic_moment + IP +
                EA + basis_set + XC_functional + length)


class Orbital():

    def __init__(self, ID, molecule_ID, line, URL='', orientation='xy', shortname=''):

        self.ID = ID
        self.database_ID = '%i.%i' % (molecule_ID, ID)
        parts = line.split(',')
        self.URL = URL + parts[0].split('=')[1]
        self.name = shortname + ' ' + parts[1].split('=')[1]
        self.energy = float(parts[2].split('=')[1])
        self.occupation = int(parts[3].split('=')[1])
        self.symmetry = parts[4].split('=')[1][:-1]
        self.orientation = orientation

    def to_string(self):

        url = 'URL: %s\n' % self.URL
        database_ID = 'Database ID: %s' % self.database_ID
        name = 'Name: %s\n' % self.name
        energy = 'Energy: %.3f\n' % self.energy
        occupation = 'Occupation: %i\n' % self.occupation
        symmetry = 'Symmetry: %s' % self.symmetry

        return url + name + energy + occupation + symmetry

    def get_meta_data(self):

        meta_data = {'name': self.name,
                     'database ID': self.database_ID,
                     'energy': self.energy,
                     'occupation': self.occupation,
                     'symmetry': self.symmetry,
                     'orientation': self.orientation}

        return meta_data



import os
import unittest
import numpy.testing as npt
import numpy as np
from kmap.library.orbital import Orbital



class TestOrbital(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        dir_path = os.path.dirname(os.path.realpath(__file__))
        filepath = dir_path + '/input/pentacene_HOMO.cube'
        with open(filepath) as file:
            cls.cubefile = file.read()
        cls.orbital = Orbital(cls.cubefile,dk3D=0.15, E_kin_max=50) 


    def setUp(self):
        pass     
        # Code run before each test comes here
        
    def tearDown(self):
        pass
        # Code run after each test comes here


    def test_init_kaxis(self):

        # check kx, ky and kz axis
        axes = ('kx', 'ky', 'kz')
        shape_expected = (49, 49, 47)
        endpoints_expected = [ [-3.6049414968815032,3.604941496881459],
                               [-3.60333247262221,  3.603332472622217],
                               [-3.4738792493962727,3.4738792493963437]]


        for axis, shape, endpoint in zip(axes, shape_expected,
                                         endpoints_expected):
            k = self.orbital.psik[axis]
            self.assertEqual(len(k), shape)  # length of axis
            npt.assert_almost_equal(k[0],  endpoint[0], decimal=13)  # first value
            npt.assert_almost_equal(k[-1], endpoint[-1],decimal=13)  # last value
 

    def test_init_3DFT(self):

        # check 3D Fourier transform
        data = self.orbital.psik['data']
        shape_expected = (49, 49, 47)
        sum_expected = 261.80992432542286
        arbitrary_element_expected = 4.352036114737106e-05  # data[11, 34, 27]

        self.assertEqual(data.shape, shape_expected)
        npt.assert_almost_equal(np.sum(data), sum_expected, decimal=13)
        npt.assert_almost_equal(data[11,34,27], arbitrary_element_expected, decimal=13)

    def test_get_kmap_basic(self):
    
        kmap = self.orbital.get_kmap(
                     E_kin=30, dk=0.10, 
                     phi=0, theta=0, psi=0,
                     Ak_type='no', polarization='p', alpha=60, beta=90,
                     gamma=0,symmetrization='no') 

        shape_expected = (56, 56)
        endpoints_expected = [-2.8060742649871617,2.8060742649871617]
        sum_expected = 11.748706872541481  

        self.assertEqual(kmap.data.shape, shape_expected)
        self.assertEqual(len(kmap.x_axis), shape_expected[0])
        self.assertEqual(len(kmap.y_axis), shape_expected[1])
        npt.assert_almost_equal(kmap.x_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.x_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(kmap.y_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.y_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)

    def test_get_kmap_orientation(self):
    
        kmap = self.orbital.get_kmap(
                     E_kin=30, dk=0.15, 
                     phi=27, theta=38, psi=-11,
                     Ak_type='no', polarization='p', alpha=60, beta=90,
                     gamma=0,symmetrization='no') 

        shape_expected = (37, 37)
        endpoints_expected = [-2.8060742649871617,2.8060742649871617]
        sum_expected = 3.9675298372410364   

        self.assertEqual(kmap.data.shape, shape_expected)
        self.assertEqual(len(kmap.x_axis), shape_expected[0])
        self.assertEqual(len(kmap.y_axis), shape_expected[1])
        npt.assert_almost_equal(kmap.x_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.x_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(kmap.y_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.y_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)


    def test_get_kmap_toroid(self):
    
        kmap = self.orbital.get_kmap(
                     E_kin=25, dk=0.20, 
                     phi=0, theta=0, psi=0,
                     Ak_type='toroid', polarization='p', alpha=60, beta=47,
                     gamma=0,symmetrization='no') 

        shape_expected = (25, 25)
        endpoints_expected = [-2.5615836216136625, 2.5615836216136625]
        sum_expected = 21.626385200246624    

        self.assertEqual(kmap.data.shape, shape_expected)
        self.assertEqual(len(kmap.x_axis), shape_expected[0])
        self.assertEqual(len(kmap.y_axis), shape_expected[1])
        npt.assert_almost_equal(kmap.x_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.x_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(kmap.y_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.y_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)

    def test_get_kmap_NanoESCA(self):
    
        kmap = self.orbital.get_kmap(
                     E_kin=25, dk=0.20, 
                     phi=0, theta=0, psi=0,
                     Ak_type='NanoESCA', polarization='p', alpha=47, beta=123,
                     gamma=0,symmetrization='no') 

        shape_expected = (25, 25)
        endpoints_expected = [-2.5615836216136625, 2.5615836216136625]
        sum_expected = 9.364599942353433   

        self.assertEqual(kmap.data.shape, shape_expected)
        self.assertEqual(len(kmap.x_axis), shape_expected[0])
        self.assertEqual(len(kmap.y_axis), shape_expected[1])
        npt.assert_almost_equal(kmap.x_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.x_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(kmap.y_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.y_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)

    def test_get_kmap_symmetry(self):
    
        kmap = self.orbital.get_kmap(
                     E_kin=25, dk=0.20, 
                     phi=12, theta=-21, psi=17,
                     Ak_type='no', polarization='p', alpha=0, beta=0,
                     gamma=0,symmetrization='3-fold+mirror') 

        shape_expected = (25, 25)
        endpoints_expected = [-2.5615836216136625, 2.5615836216136625]
        sum_expected = 3.1856202463487717   

        self.assertEqual(kmap.data.shape, shape_expected)
        self.assertEqual(len(kmap.x_axis), shape_expected[0])
        self.assertEqual(len(kmap.y_axis), shape_expected[1])
        npt.assert_almost_equal(kmap.x_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.x_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(kmap.y_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.y_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)


    def test_get_kmap_kgrid(self):
    
        kx = np.linspace(-2,2,10)
        ky = np.linspace(-2,2,12)

        kmap = self.orbital.get_kmap(
                     E_kin=25, dk=(kx,ky), 
                     phi=17, theta=-23, psi=-17,
                     Ak_type='no', polarization='p', alpha=0, beta=0,
                     gamma=0,symmetrization='no') 

        shape_expected = (10, 12)
        endpoints_expected = [-2, 2]
        sum_expected = 0.8832029249194455  

        self.assertEqual(kmap.data.T.shape, shape_expected)
        self.assertEqual(len(kmap.x_axis), shape_expected[0])
        self.assertEqual(len(kmap.y_axis), shape_expected[1])
        npt.assert_almost_equal(kmap.x_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.x_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(kmap.y_axis[0],  endpoints_expected[0], decimal=13)   
        npt.assert_almost_equal(kmap.y_axis[-1], endpoints_expected[-1],decimal=13)    
        npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)


    def test_check_new_cut(self):

        sums_expected = [202.12975607342722, 17.171157851335725,
                            3.8929271845777813,17.9086022363552,
                            27.839419217261188,1.855922052900676]

        kmaps = []
        kmaps.append(self.orbital.get_kmap(E_kin=20)) 
        kmaps.append(self.orbital.get_kmap(E_kin=20,dk=0.1))
        kmaps.append(self.orbital.get_kmap(E_kin=20,dk=0.2))
        kx = np.linspace(-2,2,50)
        ky = np.linspace(-2,2,50)
        kmaps.append(self.orbital.get_kmap(E_kin=30,dk=(kx,ky)))
        kmaps.append(self.orbital.get_kmap(E_kin=20,dk=(kx,ky)))
        kx = np.linspace(-3,3,20)
        ky = np.linspace(-3,3,20)
        kmaps.append(self.orbital.get_kmap(E_kin=20,dk=(kx,ky)))

        for kmap, sum_expected in zip(kmaps, sums_expected):
            npt.assert_almost_equal(np.nansum(kmap.data), sum_expected, decimal=13)

if __name__ == '__main__':
    unittest.main()



import unittest
import os
import numpy as np
import numpy.testing as npt
from kmap import __directory__
from kmap.library.sliceddata import SlicedData

dir_path = os.path.dirname(os.path.realpath(__file__))


class TestSlicedData(unittest.TestCase):

    def test_initialization_from_hdf5(self):

        sliced_data = SlicedData.init_from_hdf5(
            dir_path + '/../../example/data/example5_6584.hdf5')

        npt.assert_almost_equal(
            sliced_data.slice_from_index(2).data[145, 235], 194.848388671875, decimal=14)

        (sliced_data.name, '6584estep0.0170213final.txt')
        npt.assert_equal(sliced_data.meta_data, {'alias': 'M3 PTCDA/Ag(110)',
                                                 'arcwidth': '0.7800000000000011',
                                                 'fermiLevel': '28.2',
                                                 'filenumber': '6584estep0.0170213final.txt',
                                                 'kStepSize': '0.02',
                                                 'negPolar_avgs': 'False',
                                                 'polarshift': '0.0',
                                                 'rotation': '-24.0',
                                                 'sym_anglemax': '180.0',
                                                 'sym_anglemin': '0.0',
                                                 'symmode': '2-fold'})


if __name__ == '__main__':
    unittest.main()



import numpy as np


class LMFitTabModel():

    def __init__(self, sliced_data, orbitals):

        self.sliced = sliced_data
        self.orbitals = orbitals
        self.displayed_slice_data = None
        self.displayed_sum_data = None

    def get_sliced_plot(self, index, axis):

        self.displayed_slice_data = self.sliced.slice_from_index(index, axis)

        return self.displayed_slice_data

    def get_selected_orbital_plot(self, ID, parameters):

        kmap = self.get_orbital_kmap_by_ID(ID, parameters)

        return kmap

    def get_orbital_kmap_by_ID(self, ID, parameters):

        orbital = self.ID_to_orbital(ID)
        if orbital is None:
            raise IndexError('wrong ID')

        weight, *parameters = parameters

        # Get scaled kmap
        kmap = weight * orbital.get_kmap(*parameters)

        return kmap

    def get_sum_plot(self, parameters):

        kmaps = []

        for i, orbital in enumerate(self.orbitals):
            ID = orbital.ID
            kmap = self.get_orbital_kmap_by_ID(ID, parameters[i])
            kmaps.append(kmap)

        if kmaps:
            # Sum kmaps
            self.displayed_sum_data = np.nansum(kmaps)

        else:
            self.displayed_sum_data = None

        return self.displayed_sum_data

    def get_residual_plot(self):

        sum_data = self.get_sum_plot()

    def ID_to_orbital(self, ID):

        for orbital in self.orbitals:
            if orbital.ID == ID:
                return orbital

        return None



"""Defines the Orbital class.

This file defines a class named Orbital designed to read, hold,
calculate and slice data from cube files.
"""

import numpy as np
import scipy.interpolate as interp
from scipy.ndimage import rotate
from kmap.library.plotdata import PlotData
from kmap.library.misc import energy_to_k, compute_Euler_matrix

np.seterr(invalid='ignore')

class Orbital():
    """Class modelling cube files as orbitals from which kmaps can be
    sliced.

    This file defines a class named Orbital designed to read, hold,
    calculate and slice data from cube files.

    Args:
        file (string): Entire cube file to be loaded as one string.
        file_format (string): Currently only cube files are supported.
        dk3D (float): Desired resolution for 3D-Fourier-Transform.
            Single number.
        E_kin_max (float): maximum kinetic energy in eV is used to
            reduce the size of the 3D-numpy-array in momentum space
        value (string): choose between 'real', 'imag', 'abs' or 'abs2'
            for Re(), Im(), |..| or |..|^2

    Attributes:

    """

    def __init__(self, file, file_format='cube', dk3D=0.15, E_kin_max=150,value='abs2'):
        
        # Read orbital data from file
        if file_format == 'cube': 
            self._read_cube(file)
        else:
            NotImplementedError('Other formats than cube not implemented')

        self.compute_3DFT(dk3D, E_kin_max, value)
        self.kmap = {}
        self.Ak   = {}

    def get_kmap(self, E_kin=30, dk=0.03, phi=0, theta=0, psi=0,
                 Ak_type='no', polarization='p', alpha=60, beta=90,
                 gamma=0,symmetrization='no'):
        """Returns a kmap slice from the orbital data.

        Args:
            E_kin (float): Kinetic energy in eV.
            dk (float): Desired k-resolution in kmap in Angstroem^-1.
            phi (float): Euler orientation angle phi in degree.
            theta (float): Euler orientation angle theta in degree.
            psi (float): Euler orientation angle psi in degree.
            Ak_type (string): Treatment of |A.k|^2: either 'no',
                'toroid', 'NanoESCA', 'only-toroid' or 'only-NanoESCA'.
            polarization (string): Either 'p', 's', 'unpolarized', 
                'C+', 'C-' or 'CDAD'.   
            alpha (float): Angle of incidence plane in degree.
            beta (float): Azimuth of incidence plane in degree.
            gamma (float/str): Damping factor for final state in
                Angstroem^-1. str = 'auto' sets gamms automatically
            symmetrization (str): either 'no', '2-fold', '2-fold+mirror',
                '3-fold', '3-fold+mirror','4-fold', '4-fold+mirror'
        Returns:
            (PlotData): PlotData containing the kmap slice.
        """

        # Compute new hemispherical cut if E_kin or dk has changed
        new_cut = self.check_new_cut(E_kin, dk)
        if new_cut: self.set_kinetic_energy(E_kin, dk)

        # Rotate molecule (that is, rotate hemisphere) if angles have changed
        # ... and symmetrize kmap if necessary
        new_orientation = self.check_new_orientation(phi, theta, psi)       
        new_symmetrization = self.check_new_symmetrization(symmetrization)
        if new_symmetrization or new_orientation:
            self.set_orientation(phi, theta, psi)
            self.set_symmetry(symmetrization)

        # Compute polarization factor if parameters have changed
        new_Ak  = self.check_new_Ak(Ak_type, polarization, alpha, beta, gamma)
        if new_cut or new_Ak: self.set_polarization(Ak_type, polarization, 
                                         alpha, beta,gamma)

        if Ak_type == 'only-toroid' or Ak_type == 'only-NanoESCA':
            return PlotData(self.Ak['data'], self.kmap['krange'])
        
        else:
            return PlotData(self.Ak['data']*self.kmap['data'], 
                        self.kmap['krange'])


    def change_polarization(self, Ak_type='no', polarization='p', alpha=60, beta=90,
                                  gamma=0):
        
        self.set_polarization(Ak_type, polarization, alpha, beta,
                              gamma)
        return PlotData(self.Ak['data']*self.kmap['data'], 
                        self.kmap['krange'])


    def compute_3DFT(self, dk3D, E_kin_max, value):
        """Compute 3D-FT."""

        # Determine required size (nkx,nky,nkz) of 3D-FT array to reach
        # desired resolution dk3D
        pad_x = max(
            int(2 * np.pi / (dk3D * self.psi['dx'])) - self.psi['nx'], 0)
        pad_y = max(
            int(2 * np.pi / (dk3D * self.psi['dy'])) - self.psi['ny'], 0)
        pad_z = max(
            int(2 * np.pi / (dk3D * self.psi['dz'])) - self.psi['nz'], 0)

        if pad_x%2 == 1: pad_x += 1  # make sure it's an even number
        if pad_y%2 == 1: pad_y += 1  # make sure it's an even number
        if pad_z%2 == 1: pad_z += 1  # make sure it's an even number

        nkx = self.psi['nx'] + pad_x
        nky = self.psi['ny'] + pad_y
        nkz = self.psi['nz'] + pad_z

        # Set up k-grids for 3D-FT
        kx = self.set_3Dkgrid(nkx, self.psi['dx'])
        ky = self.set_3Dkgrid(nky, self.psi['dy'])
        kz = self.set_3Dkgrid(nkz, self.psi['dz'])

        # Compute 3D FFT
        # !! TESTING !!! This is supposed to yield also proper Real- and Imaginary parts!!
#        print(nkx, nky, nkz)
#        print(pad_x, pad_y, pad_z)
        psi_padded = np.pad(self.psi['data'], 
                      pad_width=((pad_x//2,pad_x//2), (pad_y//2,pad_y//2), (pad_z//2,pad_z//2)),
                      mode='constant',
                      constant_values=(0,0))
        psi_padded = np.fft.ifftshift(psi_padded)                                
        psik = np.fft.fftshift(np.fft.fftn(psi_padded))

       # THIS IS THE OLD AND WELL TESTED WAY TO Compute 3D FFT
#        psik = np.fft.fftshift(np.fft.fftn(self.psi['data'],
#                                           s=[nkx, nky, nkz]))

        # properly normalize wave function in momentum space
        dkx, dky, dkz = kx[1]-kx[0], ky[1]-ky[0], kz[1]-kz[0]
        factor        = dkx*dky*dkz*np.sum(np.abs(psik)**2)
        psik         /= np.sqrt(factor)

        # Reduce size of array to value given by E_kin_max to save memory
        k_max      = energy_to_k(E_kin_max)
        kx_indices = np.where((kx <= k_max) & (kx >= -k_max))[0]
        ky_indices = np.where((ky <= k_max) & (ky >= -k_max))[0]
        kz_indices = np.where((kz <= k_max) & (kz >= -k_max))[0]    
        kx         = kx[kx_indices]
        ky         = ky[ky_indices]
        kz         = kz[kz_indices]
        psik       = np.take(psik, kx_indices, axis=0)
        psik       = np.take(psik, ky_indices, axis=1)
        psik       = np.take(psik, kz_indices, axis=2)

        # decide whether real, imaginry part, absolute value, or squared absolute value is used
        if value == 'real':
            psik = np.asarray(np.real(psik), order='C')

        elif value == 'imag':
            psik = np.asarray(np.imag(psik), order='C')

        elif value == 'abs':
            psik = np.abs(psik)

        else:
            psik = np.abs(psik)**2


        # Define interpolating function to be used later for kmap
        # computation
        psik_interp = interp.RegularGridInterpolator((kx, ky, kz), psik,
                                                     bounds_error=False,
                                                     fill_value=np.nan)

        # Set attributes
        self.psik = {'kx': kx, 'ky': ky, 'kz': kz,
                     'E_kin_max':E_kin_max,
                     'value': value,
                     'data': psik,
                     'data_interp': psik_interp}

        # Why? Local variables should only live until end of function
        # anyways...
        # Free memory for psik-array
        del psik

    # Make hemi-spherical cut through 3D Fourier transform
    def set_kinetic_energy(self, E_kin, dk):

        kmax = energy_to_k(E_kin)
        if type(dk) == tuple:
            kxi = dk[0]
            kyi = dk[1]
            num_kx = len(kxi)
            num_ky = len(kyi)
        else:
            num_kx = int(2 * kmax / dk)
            num_ky = int(2 * kmax / dk)
            kxi = np.linspace(-kmax, +kmax, num_kx)
            kyi = np.linspace(-kmax, +kmax, num_ky)

        krange = ((kxi[0], kxi[-1]), (kyi[0], kyi[-1]))
        KX, KY = np.meshgrid(kxi, kyi, indexing='xy')
        KZ = np.sqrt(kmax**2 - KX**2 - KY**2)
        kxkykz = list(map(lambda a, b, c: (a, b, c),
                          KX.flatten(), KY.flatten(), KZ.flatten()))
        data = np.reshape(self.psik['data_interp'](kxkykz), (num_kx, num_ky))

        # Set kmap attributes
        self.kmap = {'E_kin': E_kin, 'dk': dk, 'krange': krange,
                     'KX': KX, 'KY': KY, 'KZ': KZ,
                     'phi': 0, 'theta': 0, 'psi': 0,
                     'data': data}

    def check_new_cut(self, E_kin, dk):

        eps = 1e-10

        new_cut = False

        if 'E_kin' in self.kmap:

            if type(dk) != tuple and type(self.kmap['dk']) != tuple: 
                if (np.abs(self.kmap['E_kin'] - E_kin) > eps or 
                    np.abs(self.kmap['dk']    - dk)    > eps):
                   new_cut = True
           
            elif type(dk) == tuple and type(self.kmap['dk']) != tuple:
                new_cut = True

            elif type(dk) != tuple and type(self.kmap['dk']) == tuple:
                new_cut = True

            elif type(dk) == tuple and type(self.kmap['dk']) == tuple:
                if np.abs(self.kmap['E_kin'] - E_kin) > eps:
                    new_cut = True

                if dk[0].shape != self.kmap['dk'][0].shape:
                    new_cut = True
                else:
                    if np.any( np.abs(self.kmap['dk'][0] - dk[0]) > eps):
                        new_cut = True 

                if dk[1].shape != self.kmap['dk'][1].shape:
                    new_cut = True
                else:
                    if np.any( np.abs(self.kmap['dk'][1] - dk[1]) > eps):
                        new_cut = True 

        else:
            new_cut = True

        return new_cut


    # get the (kx,ky) values of the kmap as a list of tuples 
    def get_kxkygrid(self):
        
        KX, KY = self.kmap['KX'], self.kmap['KY']   
        return list(map(lambda a, b: (a, b), KX.flatten(), KY.flatten()))         

    # Rotate hemisphere KX, KY, KZ by Euler angles phi, theta, psi
    def set_orientation(self, phi, theta, psi):

        KX, KY, KZ = self.kmap['KX'], self.kmap['KY'], self.kmap['KZ']
        nkx = KX.shape[0]
        nky = KX.shape[1]
        r = compute_Euler_matrix(phi, theta, psi)
        r = r.T

        # KXr, KYr, KZr are the k-coordinates of the rotated hemisphere
        KXr = r[0, 0] * KX + r[0, 1] * KY + r[0, 2] * KZ
        KYr = r[1, 0] * KX + r[1, 1] * KY + r[1, 2] * KZ
        KZr = r[2, 0] * KX + r[2, 1] * KY + r[2, 2] * KZ
        kxkykz = list(map(lambda a, b, c: (a, b, c),
                          KXr.flatten(), KYr.flatten(), KZr.flatten()))
        data = np.reshape(self.psik['data_interp'](kxkykz), (nkx, nky))

        # update attributes
        self.kmap['phi'] = phi
        self.kmap['theta'] = theta
        self.kmap['psi'] = psi
        self.kmap['data'] = data

    def check_new_orientation(self, phi, theta, psi):

        eps = 1e-10
        if 'phi' in self.kmap:
            if (np.abs(self.kmap['phi']   - phi)   > eps or 
                np.abs(self.kmap['theta'] - theta) > eps or
                np.abs(self.kmap['psi']   - psi)   > eps ):
                  new_orientation = True

            else:
                  new_orientation = False
        else:
            new_orientation = True

        return new_orientation


    def set_polarization(self, Ak_type, polarization, alpha, beta, gamma):

        if Ak_type == 'no':  # Set |A.k|^2 to 1
            self.Ak = {'Ak_type': Ak_type,
                       'polarization': polarization,
                       'alpha': alpha,
                       'beta': beta,
                       'gamma': gamma,
                       'data': np.ones_like(self.kmap['data'])}

            return


        # Convert angles to rad and compute sin and cos for later use
        a = np.radians(alpha)
        b = np.radians(beta)
        sin_a = np.sin(a)
        cos_a = np.cos(a)
        sin_b = np.sin(b)
        cos_b = np.cos(b)

        # Compute gamma according to inelastic free mean path
        if gamma == 'auto':
            # lambda is calculated from the "universal curve" empirical
            # relation
            E_kin = self.kmap['E_kin']
            c1 = 143
            c2 = 0.054
            lam = c1 * E_kin**(-2) + c2 * np.sqrt(E_kin)
            lam *= 10
            gamma_calc = 1 / lam
        else:
            gamma_calc = gamma

        # Retrieve k-grid from kmap
        kx, ky, kz = self.kmap['KX'], self.kmap['KY'], self.kmap['KZ']
        # Magnitude of k-vector
        k2 = kx**2 + ky**2 + kz**2
        kmax = energy_to_k(self.kmap['E_kin'])

        # At the toroid, the emitted electron is always in the plane of
        # incidence and the sample is rotated
        if Ak_type == 'toroid' or Ak_type == 'only-toroid':     
            # Parallel component of k-vector
            kpar = np.sqrt(k2 - kz**2)
            # |A.k|^2 factor
            Ak = (kpar * cos_a + kz * sin_a)**2

        # At the NanoESCA, either p-polarization ,s-polarization, or
        # circularly polarized light can be simulated
        elif Ak_type == 'NanoESCA' or Ak_type == 'only-NanoESCA':
            # In-plane = p-polarization
            if polarization == 'p':
                Ak = kx * cos_a * cos_b + ky * cos_a * sin_b + kz * sin_a
                Ak = Ak**2 + gamma_calc**2 * sin_a**2

            # Out-of-plane = s-polarization
            elif polarization == 's':
                Ak = -kx * sin_b + ky * cos_b
                Ak = Ak**2
                Ak[kx**2 + ky**2 > kmax**2] = np.nan


            # unpolarized light, e.g. He-lamp 
            # Compare Equation (37) in S. Moser, J. Electr. Spectr.
            # Rel. Phen. 214, 29-52 (2017).            
            elif polarization == 'unpolarized':
                Ak = (k2 + gamma_calc**2 + 2*kx*kz*cos_b + 2*ky*kz*sin_b)*np.sin(2*a)  
                Ak+= (kx**2*sin_b + ky**2*cos_b - kz**2 - gamma_calc**2)*np.cos(2*a)  
                Ak*= (2/3)       

            # Circularly polarized light (right-handed)
            elif polarization == 'C+':
                polp = kx * cos_a * cos_b + ky * cos_a * sin_b + kz * sin_a
                pols = -kx * sin_b + ky * cos_b
                Ak = 0.5 * (polp**2 + gamma_calc**2 * sin_a**2) + 0.5 * \
                    pols**2 + (sin_b * kx - cos_b * ky) * gamma_calc * sin_a

            # Circularly polarized light (left-handed)
            elif polarization == 'C-':
                polp = kx * cos_a * cos_b + ky * cos_a * sin_b + kz * sin_a
                pols = -kx * sin_b + ky * cos_b
                Ak = 0.5 * (polp**2 + gamma_calc**2 * sin_a**2) + 0.5 * \
                    pols**2 - (sin_b * kx - cos_b * ky) * gamma_calc * sin_a

            # CDAD-signal (right-handed - left-handed) using empirically
            # damped plane wave
            elif polarization == 'CDAD':
                # Compare Equation (31) in S. Moser, J. Electr. Spectr.
                # Rel. Phen. 214, 29-52 (2017).
                Ak = +2 * (sin_b * kx - cos_b * ky) * gamma_calc * sin_a
                Ak[kx**2 + ky**2 > kmax**2] = np.nan

        # Set attributes
        self.Ak = {'Ak_type': Ak_type,
                   'polarization': polarization,
                   'alpha': alpha,
                   'beta': beta,
                   'gamma': gamma,
                   'data': Ak}

    def check_new_Ak(self, Ak_type, polarization, alpha, beta, gamma):

        if 'Ak_type' in self.Ak:

            if (self.Ak['Ak_type']      != Ak_type      or
                self.Ak['polarization'] != polarization or
                self.Ak['alpha']        != alpha        or
                self.Ak['beta']         != beta         or
                self.Ak['gamma']        != gamma):
 
                  new_Ak = True

            else:
                  new_Ak = False
        else:
            new_Ak = True

        return new_Ak


    def set_symmetry(self, symmetrization):
        """Symmterizes the kmap.

        Args:
            symmetrization (str): either 'no', '2-fold', '2-fold+mirror',
                '3-fold', '3-fold+mirror','4-fold', '4-fold+mirror'

        """
        if symmetrization == 'no':
            self.kmap['symmetrization'] = 'no'
            return 

        data = self.kmap['data']

        if symmetrization == '2-fold': 
            data += rotate(np.nan_to_num(data), 180, reshape=False)
            data /= 2
                    
        elif symmetrization == '2-fold+mirror': 
            data += rotate(np.nan_to_num(data), 180, reshape=False)
            data += np.flip(data, 0)  # mirror map with respect to first axis
            data /= 4

        elif symmetrization == '3-fold': 
            data1 = rotate(np.nan_to_num(data), 120, reshape=False)
            data2 = rotate(np.nan_to_num(data), 240, reshape=False)
            data += (data1 + data2)
            data /= 3
                     
        elif symmetrization == '3-fold+mirror': 
            data1 = rotate(np.nan_to_num(data), 120, reshape=False)
            data2 = rotate(np.nan_to_num(data), 240, reshape=False)
            data += (data1 + data2)
            data += np.flip(data, 0)
            data /= 6

        elif symmetrization == '4-fold': 
            data1 = rotate(np.nan_to_num(data), 90, reshape=False)
            data2 = rotate(np.nan_to_num(data), 180, reshape=False)
            data3 = rotate(np.nan_to_num(data), 270, reshape=False)
            data += (data1 + data2 + data3)
            data /= 4
                      
        elif symmetrization == '4-fold+mirror': 
            data1 = rotate(np.nan_to_num(data), 90, reshape=False)
            data2 = rotate(np.nan_to_num(data), 180, reshape=False)
            data3 = rotate(np.nan_to_num(data), 270, reshape=False)
            data += (data1 + data2 + data3)
            data += np.flip(data, 0)  # mirror map with respect to first axis
            data += np.flip(data, 1)  # mirror map with respect to second axis
            data /= 8

        self.kmap['symmetrization'] = symmetrization
        self.kmap['data'] = data
        return

    def check_new_symmetrization(self, symmetrization):

        if 'symmetrization' in self.kmap:

            if self.kmap['symmetrization'] != symmetrization:
 
                  new_symmetrization = True

            else:
                  new_symmetrization = False
        else:
            new_symmetrization = True

        return new_symmetrization


    def plot(self, ax, title=None, kxlim=None, kylim=None, interpolation='bicubic'):
        """Creates a plot of the kmap in axes-obeject ax.

        Args:
            

        """

        # prepare data
        data   = self.Ak['data']*self.kmap['data']  
        krange = self.kmap['krange']
        limits = [krange[0][0], krange[0][1], krange[1][0], krange[1][1]]

        # plot kmap       
        im = ax.imshow(data,
                      extent=limits,
                      interpolation=interpolation,
                      origin='lower',
                      cmap='jet')

        # format plot
        ax.set_aspect('equal')
        ax.set_xlabel(r'$\kappa_x(1/\AA)$')
        ax.set_ylabel(r'$\kappa_y(1/\AA)$')

        if kxlim != None:
            ax.set_xlim(kxlim[0],kxlim[1])        
        if kylim != None:
            ax.set_ylim(kxlim[0],kxlim[1])        

        if title != None:
            ax.set_title(title)

        return

    def _read_cube(self, file):
        """ Read orbital data from cube file."""

        ''' File and path to file are both string. Should _read_cube
        get the file path or the file itself?
        if type(file) == str:
            orbfile = open(file, "r")
            lines   = orbfile.readlines()
            orbfile.close()
        else:
        '''
        # Split the entire file into separate lines
        lines = file.split('\n')

        # Set up real space grid info
        # Conversion factor from Bohr to Angstroem
        b2a = 0.529177105787531

        name = lines[1].strip()

        x0, y0, z0 = b2a * float(lines[2].split()[1]), b2a * float(
            lines[2].split()[2]), b2a * float(lines[2].split()[3])
        nx, ny, nz = int(lines[3].split()[0]), int(
            lines[4].split()[0]), int(lines[5].split()[0])
        dx, dy, dz = b2a * float(lines[3].split()[1]), b2a * float(
            lines[4].split()[2]), b2a * float(lines[5].split()[3])
        lx, ly, lz = (nx - 1) * dx, (ny - 1) * dy, (nz - 1) * dz
        x = np.linspace(x0, x0 + lx, nx)
        y = np.linspace(y0, y0 + ly, ny)
        z = np.linspace(z0, z0 + lz, nz)

        # Read atomic coordinates
        # Number of atoms from line 3
        num_atom = int(lines[2].split()[0])
        chemical_numbers = []
        atomic_coordinates = []
        for line in lines[6:6 + num_atom]:
            words = line.split()
            chemical_numbers.append(int(words[0]))
            a, b, c = float(words[2]), float(words[3]), float(words[4])
            atomic_coordinates.append([b2a * a, b2a * b, b2a * c])

        atomic_coordinates = np.array(atomic_coordinates)

        # Now read cube data
        data = []
        for line in lines[(6 + num_atom):]:
            words = line.split()
            for word in words:
                data.append(float(word))
        # Conversion to 3D-numpy array
        data = np.reshape(np.array(data), (nx, ny, nz))

        # set attributes
        self.psi = {'name': name,
                    'nx': nx, 'ny': ny, 'nz': nz,
                    'dx': dx, 'dy': dy, 'dz': dz,
                    'x': x, 'y': y, 'z': z,
                    'data': data}
        self.molecule = {'num_atom': num_atom,
                         'chemical_numbers': chemical_numbers,
                         'atomic_coordinates': atomic_coordinates}

    def get_bonds(self,lower_factor=0.8,upper_factor=1.2):
        """ returns a list of bond used for plotting the molecular structure.

        Args:
            lower_factor (float): lower bound for drawing bonds w.r.t sum of covalent radii
            upper_factor (float): upper bound for drawing bonds w.r.t sum of covalent radii
        """
        covalent_R = {1:0.32, 2:0.32,  # H, He
                      3:1.34, 4:0.90, 5:0.82, 6:0.77, 7:0.71, 8:0.73, 9:0.71,10:0.69, # Li - Ne
                     11:1.54,12:1.30,13:1.18,14:1.11,15:1.06,16:1.02,17:0.99,18:0.97, # Na- Ar
                     19:1.96,20:1.74,21:1.44,22:1.36,23:1.25,24:1.27,25:1.39,26:1.25, # K - Fe
                     27:1.26,28:1.21,29:1.38,30:1.31,31:1.26,32:1.22,33:1.21,34:1.16, # Co- Se
                     35:1.14,36:1.10,                                                 # Br, Kr
                     37:2.11,38:1.92,39:1.62,40:1.48,41:1.37,41:1.45,43:1.31,44:1.26, # Rb -Ru
                     45:1.35,46:1.31,47:1.53,48:1.48,49:1.44,50:1.41,51:1.38,52:1.35, # Rh -Te
                     53:1.33,54:1.30,                                                 #  I, Xe
                     55:2.25,56:1.98,57:1.69,72:1.50,73:1.38,74:1.46,75:1.59,76:1.28, # Cs -Os
                     77:1.37,78:1.38,79:1.38,80:1.49,81:1.48,82:1.46,83:1.46,84:1.40, # Ir -Po
                     85:1.45,86:1.45}                                                 # At, Rn
               
        dx,dy,dz    = self.psi['dx'], self.psi['dy'], self.psi['dz'] 
        coordinates = self.molecule['atomic_coordinates']
        Z_list      = self.molecule['chemical_numbers']                  
        bonds       = []
        for atom1, Z1 in zip(coordinates, Z_list):
            x1,y1,z1 = atom1[0], atom1[1], atom1[2]
            R1       = covalent_R[Z1]
            for atom2, Z2 in zip(coordinates, Z_list):
                x2,y2,z2 = atom2[0], atom2[1], atom2[2] 
                R2       = covalent_R[Z2] 
                R        = R1 + R2    # sum of covalent radii 
                distance = np.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)
                if lower_factor*R <= distance <= upper_factor*R:
                    bond = [[x1/dx,y1/dy,z1/dz], [x2/dx,y2/dy,z2/dz]] 
                    bonds.append(np.array(bond))

        return bonds


    def set_3Dkgrid(self, nk, delta):

        k_min = -np.pi/delta
        k_max = +np.pi/delta
        k, dk = np.linspace(k_min, k_max, nk, retstep=True)
        if nk%2 == 0:
            k = k - dk/2

        return k





# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import Qt, pyqtSignal, QDir
from PyQt5.QtWidgets import QMainWindow, QTreeWidgetItem, QHeaderView

# Own Imports
from kmap import __directory__
from kmap.library.database import Database
from kmap.controller.sliceddatabaseoptions import SlicedDataBaseOptions
from kmap.controller.sliceddatabaseoptions2 import SlicedDataBaseOptions2
from kmap.controller.slicedcubefileoptions import SlicedCubefileOptions
from kmap.config.config import config

# Load .ui File for SlicedDataDatabase
UI_file = __directory__ + QDir.toNativeSeparators('/ui/databasewindow.ui')
DatabaseWindow_UI, _ = uic.loadUiType(UI_file)
# Load .ui File for OrbitalDatabase
UI_file = __directory__ + \
    QDir.toNativeSeparators('/ui/sliceddatabasewindow.ui')
SlicedDatabaseWindow_UI, _ = uic.loadUiType(UI_file)


class DatabaseWindowBase(QMainWindow):

    files_chosen = pyqtSignal(list)

    def __init__(self):

        # Setup GUI
        super(DatabaseWindowBase, self).__init__()
        self.setAttribute(Qt.WA_DeleteOnClose, True)

    def fill_tree(self):

        self.tree.clear()

        for molecule in self.database.molecules:
            if self._fit_filter(molecule):
                item = self.add_molecule(molecule)

    def add_molecule(self, molecule):

        molecule_item = MoleculeTreeWidgetItem(molecule)

        # Add the entries in the respective column
        entries = ['%i' % molecule.ID,
                   molecule.full_name,
                   molecule.formula,
                   '%i' % molecule.charge,
                   '%.3f' % molecule.magnetic_moment,
                   molecule.XC_functional,
                   '']

        for col, entry in enumerate(entries):
            molecule_item.setText(col, entry)
            molecule_item.setTextAlignment(col, Qt.AlignCenter)

        # Add item to tree
        self.tree.addTopLevelItem(molecule_item)

        # Add orbitals of molecule to tree
        for orbital in molecule.orbitals:
            self.add_orbital(orbital, molecule_item)

    def add_orbital(self, orbital, molecule_item):

        orbital_item = OrbitalTreeWidgetItem(orbital)

        # Add the entries in the respective column
        entries = ['%i' % orbital.ID, orbital.name, '', '',
                   '', '', '%.3f' % orbital.energy]

        for col, entry in enumerate(entries):
            orbital_item.setText(col, entry)
            orbital_item.setTextAlignment(col, Qt.AlignCenter)

        # Add item to tree
        molecule_item.addChild(orbital_item)

    def find(self):

        search_text = self.line_edit.text()

        if not search_text:
            return

        items = self.tree.findItems(search_text,
                                    Qt.MatchFlag.MatchContains, 1)

        if not items:
            return

        current_item = self.tree.currentItem()
        if current_item in items:
            current_index = items.index(current_item)
            next_index = (current_index + 1) % len(items)

        else:
            next_index = 0

        next_item = items[next_index]
        self.tree.setCurrentItem(next_item)

    def load_database(self):

        orbitals_chosen = self._get_chosen_orbitals()

        for orbital in orbitals_chosen:
            URL = orbital.URL
            meta_data = orbital.get_meta_data()
            self.URLs.append([URL, meta_data])

        self.close()

    def closeEvent(self, event):

        self.files_chosen.emit(self.URLs)
        self.deleteLater()
        event.accept()

    def _setup_tree(self):

        header_labels = ['ID', 'Name', 'Formula', 'Charge',
                         'Mag. Moment', 'XC-Funktional', 'Energy']
        self.tree.setHeaderLabels(header_labels)
        self.tree.setColumnWidth(0, 100)
        self.tree.setColumnWidth(2, 150)
        self.tree.setColumnWidth(3, 100)
        self.tree.setColumnWidth(4, 150)
        self.tree.setColumnWidth(5, 150)
        self.tree.setColumnWidth(6, 120)
        self.tree.header().setDefaultAlignment(Qt.AlignCenter)
        self.tree.header().setSectionResizeMode(1, QHeaderView.Stretch)

        self.tree.sortItems(0, Qt.SortOrder.AscendingOrder)

    def _fit_filter(self, molecule):

        filter_ = self.combobox.currentText()

        if filter_ == 'No Filter'or molecule.XC_functional == filter_:
            return True

        else:
            return False

    def _get_chosen_orbitals(self):

        orbitals = []

        selected_items = self.tree.selectedItems()

        for selected_item in selected_items:
            # Top level items have no parent and are molecules
            if not selected_item.parent():
                molecule = selected_item.molecule

                for orbital in molecule.orbitals:
                    orbitals.append(orbital)

            else:
                orbital = selected_item.orbital
                orbitals.append(orbital)

        # Set removes duplicates from selecting molecule and orbitals
        return set(orbitals)

    def _connect(self):

        self.combobox.currentTextChanged.connect(self.fill_tree)

        self.database_load_button.clicked.connect(self.load_database)
        self.find_button.clicked.connect(self.find)


class OrbitalDatabase(DatabaseWindowBase, DatabaseWindow_UI):

    def __init__(self, *args, **kwargs):

        super(OrbitalDatabase, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup_tree()
        self._connect()

        # Setup database
        path = __directory__ + config.get_key('paths', 'database')
        self.database = Database(path)

        # URLs (with extra information if available) chosen
        self.URLs = []

        self.fill_tree()

        self.show()

    def load_urls(self):

        text = self.text_edit.toPlainText()

        # Nothing entered
        if not text:
            self.URLs = []

        # Multiple URLs entered
        elif '\n' in text:
            self.URLs = text.split('\n')

        # Only one URL entered
        else:
            self.URLs.append([text, {}])

        self.close()

    def _connect(self):

        DatabaseWindowBase._connect(self)

        self.url_load_button.clicked.connect(self.load_urls)


class SlicedDatabaseWindow(DatabaseWindowBase, SlicedDatabaseWindow_UI):

    def __init__(self, mode, *args, **kwargs):

        super(SlicedDatabaseWindow, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup_tree()
        self._connect()

        # Setup database
        path = __directory__ + config.get_key('paths', 'database')
        self.database = Database(path)

        # Open Options Window
        if mode == 'binding-energy':
            self.options = SlicedDataBaseOptions()

        elif mode == 'photon-energy':
            self.options = SlicedDataBaseOptions2()

        elif mode == 'cubefile':
            self.options = SlicedCubefileOptions()

        # URLs (with extra information if available) chosen
        self.URLs = []

        self.fill_tree()

        self.show()
        self.options.show()

    def load_database(self):

        orbitals_chosen = self._get_chosen_orbitals()

        for orbital in orbitals_chosen:
            URL = orbital.URL
            meta_data = orbital.get_meta_data()
            options = self.options.get_parameters()
            self.URLs.append([URL, meta_data])

        self.URLs.append(options)

        self.close()

    def closeEvent(self, event):

        DatabaseWindowBase.closeEvent(self, event)

        try:
            self.options.deleteLater()

        except RuntimeError:
            pass


class TreeWidgetItem(QTreeWidgetItem):

    def __lt__(self, other):

        column = self.treeWidget().sortColumn()

        self_string = self.text(column).lower()
        other_string = other.text(column).lower()

        if self_string.isnumeric() and other_string.isnumeric():
            return float(self_string) < float(other_string)

        else:
            return self_string < other_string


class MoleculeTreeWidgetItem(TreeWidgetItem):

    def __init__(self, molecule, *args, **kwargs):

        self.molecule = molecule

        super(MoleculeTreeWidgetItem, self).__init__(*args, **kwargs)


class OrbitalTreeWidgetItem(TreeWidgetItem):

    def __init__(self, orbital, *args, **kwargs):

        self.orbital = orbital

        super(OrbitalTreeWidgetItem, self).__init__(*args, **kwargs)



# PyQt5 Imports
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget, QPushButton

# Own Imports
from kmap.library.misc import orientation_to_euler_angle
from kmap.library.qwidgetsub import (CenteredLabel, UseCheckBox,
                                     AngleSpinBox, WeightSpinBox)


class OrbitalTableRow(QWidget):

    row_removed = pyqtSignal(int)
    parameter_changed = pyqtSignal(int, str, float)
    use_changed = pyqtSignal(int)

    def __init__(self, orbital, orientation, *args, **kwargs):

        self.ID = orbital.ID

        super(OrbitalTableRow, self).__init__(*args, **kwargs)

        self._setup(orbital, orientation)
        self._connect()

    def remove_row(self):

        self.row_removed.emit(self.ID)

    def get_parameters(self):

        weight = self.weight.value()
        phi = self.phi.value()
        theta = self.theta.value()
        psi = self.psi.value()

        return weight, phi, theta, psi

    def get_use(self):

        return self.use.isChecked()

    def change_parameter(self):

        sender = self.sender()

        parameter = sender.objectName()
        value = sender.value()

        self.parameter_changed.emit(self.ID, parameter, value)

    def change_use(self):

        self.use_changed.emit(self.ID)

    def update_parameter_silently(self, parameter, value):

        widgets = [self.weight, self.phi, self.theta, self.psi]

        for widget in widgets:
            if widget.objectName() == parameter:

                widget.blockSignals(True)
                widget.setValue(value)
                widget.blockSignals(False)

    def _setup(self, orbital, orientation):

        self.button = QPushButton('X')
        self.ID_label = CenteredLabel('%i' % orbital.ID)
        self.name_label = CenteredLabel(orbital.name)
        self.weight = WeightSpinBox()

        phi, theta, psi = orientation_to_euler_angle(orientation)
        self.phi = AngleSpinBox(phi, 'phi')
        self.theta = AngleSpinBox(theta, 'theta')
        self.psi = AngleSpinBox(psi, 'psi')

        self.use = UseCheckBox()

        self.name_label.setToolTip(str(orbital))

    def _connect(self):

        self.button.clicked.connect(self.remove_row)

        self.weight.valueChanged.connect(self.change_parameter)
        self.phi.valueChanged.connect(self.change_parameter)
        self.theta.valueChanged.connect(self.change_parameter)
        self.psi.valueChanged.connect(self.change_parameter)

        self.use.stateChanged.connect(self.change_use)



import json


class Colormap():

    def __init__(self, name, pos, colors):

        self.name = name
        self.pos = pos
        self.colors = colors

    def toList(self):

        return [self.name, self.pos, self.colors]



# Python Imports
import numpy as np
import pyqtgraph as pg

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import QDir

# Own Imports
from kmap import __directory__
from kmap.config.config import config
from kmap.library.misc import normalize
from kmap.model.crosshair_model import CrosshairROIModel
from kmap.controller.crosshair import CrosshairBase

# Load .ui File
UI_file = __directory__ + QDir.toNativeSeparators('/ui/crosshairroi.ui')
CrosshairROI_UI, _ = uic.loadUiType(UI_file)


class CrosshairROIBase(CrosshairBase):

    def __init__(self, plot_item):

        super().__init__(plot_item)

    def enable_roi(self, enable):

        self.roi.setVisible(enable)

    def dragging_roi(self):

        self.dragging_roi = True

    def resize_roi_from_drag(self):

        # There is no signal for finished dragging, but starting it
        if not self.dragging_roi:
            return

        self.dragging_roi = False

        radius = self.roi.size()[0] / 2

        if radius <= 0:
            radius = 0.01

        self.model.set_radius(radius)

        self.update()

    def resize_roi_from_spinbox(self):

        radius = self.roi_spinbox.value()

        if radius <= 0:
            radius = 0.01

        self.model.set_radius(radius)

        self.update()

    def move_crosshair_from_spinbox(self):

        update = super().move_crosshair_from_spinbox()

        if update:
            x, y = self.model.x, self.model.y
            self.roi.setPos(x, y)

            return True

        else:
            return False

    def change_color(self, index):

        CrosshairBase.change_color(self, index)

        self.roi.setPen(self.colors[index])

    def update(self):

        x, y, radius = self.model.x, self.model.y, self.model.radius

        # Update Spinboxes silently
        self.roi_spinbox.blockSignals(True)
        self.roi_spinbox.setValue(radius)
        self.roi_spinbox.blockSignals(False)

        # Update Crosshair Position and Size
        self.roi.setPos([x - radius, y - radius])
        self.roi.setSize([2 * radius, 2 * radius])

        super().update()

    def update_label(self):

        super().update_label()

        plot_data = self.plot_item.get_plot_data()

        if plot_data == None:
            intensity = 0

        else:
            cut = self.model.cut_from_data(plot_data, region='roi')

            # Normalize by dividing by the number of non nan elements
            if config.get_key('crosshair', 'normalized_intensity') == 'True':
                intensity = normalize(cut.data)

            else:
                intensity = np.nansum(cut.data)

        if abs(intensity) > 1000:
            self.area_value_label.setText('%.2fk' % (intensity / 1000))

        else:
            self.area_value_label.setText('%.2f' % intensity)

    def _set_model(self, model=None):

        if model is None:
            self.model = CrosshairROIModel(x=0, y=0, radius=0.02)

        else:
            self.model = model

    def _setup(self):

        CrosshairBase._setup(self)

        x, y, radius = self.model.x, self.model.y, self.model.radius

        self.roi = pg.CircleROI([x - radius, y - radius],
                                size=[2 * radius, 2 * radius],
                                movable=False,
                                rotatable=False,
                                resizable=True,
                                removable=False,
                                pen='k',
                                handlePen='r')
        self.plot_item.addItem(self.roi)

    def _connect(self):

        CrosshairBase._connect(self)

        self.enable_roi_checkbox.stateChanged.connect(self.enable_roi)

        self.roi_spinbox.valueChanged.connect(self.resize_roi_from_spinbox)
        self.roi.sigRegionChangeFinished.connect(self.resize_roi_from_drag)
        self.roi.sigRegionChangeStarted.connect(self.dragging_roi)


class CrosshairROI(CrosshairROIBase, CrosshairROI_UI):

    def __init__(self, plot_item):

        # Setup GUI
        super(CrosshairROI, self).__init__(plot_item)
        self.setupUi(self)
        self._setup()
        self._connect()

        self.dragging_roi = False
        self.enable_roi(False)

        self.enable(False)
        self.update_label()



import os
import unittest
import numpy as np
import numpy.testing as npt
from kmap.library.plotdata import PlotData
from kmap.model.crosshair_model import *


dir_path = os.path.dirname(os.path.realpath(__file__))


class TestCrosshair(unittest.TestCase):

    @classmethod
    def setUpClass(self):

        self.expected = np.load(
            dir_path + '/output/test_crosshair_expected.npy',
            allow_pickle=True).item()

    def setUp(self):

        self.x = 0
        self.y = 0
        self.radius = 1.5
        self.width = 1.5
        self.crosshair = CrosshairAnnulusModel(x=self.x, y=self.y,
                                               radius=self.radius,
                                               width=self.width)

        self.data = PlotData(np.reshape(np.array(range(99)), (9, 11)),
                             [[-5, 5], [-4, 4]])

    def test_correct_initialization(self):

        self.assertEqual(self.crosshair.x, self.x)
        self.assertEqual(self.crosshair.y, self.y)
        self.assertEqual(self.crosshair.radius, self.radius)
        self.assertEqual(self.crosshair.width, self.width)

    def test_incorrect_initialization(self):

        self.assertRaises(ValueError, CrosshairROIModel, radius=-1)
        self.assertRaises(ValueError, CrosshairAnnulusModel, width=-1)

    def test_set_position(self):

        new_x = 1
        new_y = -1

        self.crosshair.set_position()
        self.assertEqual(self.crosshair.x, self.x)
        self.assertEqual(self.crosshair.y, self.y)

        self.crosshair.set_position(y=new_y)
        self.assertEqual(self.crosshair.x, self.x)
        self.assertEqual(self.crosshair.y, new_y)

        self.crosshair.set_position(new_x)
        self.assertEqual(self.crosshair.x, new_x)
        self.assertEqual(self.crosshair.y, new_y)

    def test_set_radius(self):

        new_radius = 2
        self.crosshair.set_radius(new_radius)
        self.assertEqual(self.crosshair.radius, new_radius)

        new_radius = 0
        self.crosshair.set_radius(new_radius)
        self.assertEqual(self.crosshair.radius, new_radius)

        new_radius = -1
        self.assertRaises(ValueError, self.crosshair.set_radius, new_radius)

    def test_set_width(self):

        new_width = 2
        self.crosshair.set_width(new_width)
        self.assertEqual(self.crosshair.width, new_width)

        new_width = 0
        self.crosshair.set_width(new_width)
        self.assertEqual(self.crosshair.width, new_width)

        new_width = -1
        self.assertRaises(ValueError, self.crosshair.set_width, new_width)

    def test_point_mask(self):

        # Standard
        title = 'Center Cut'
        self.mask(1, 1, 2, 1.5, 'center', False, title)

        # Edge Case
        title = 'Center Edge Case'
        self.mask(0.5, 0.5, 1.5, 1.5, 'center', False, title)

    def test_line_mask(self):

        # x-Line
        title = 'x-Line'
        self.mask(0, 0, 1.5, 1.5, 'x', False, title)

        # y-Line
        title = 'y-Line'
        self.mask(0, 0, 1.5, 1.5, 'y', False, title)

        # x-Line Inverted
        title = 'x-Line Inverted'
        self.mask(0, 0, 1.5, 1.5, 'x', True, title)

        # x-Line 2
        title = 'x-Line 2'
        self.mask(1, 0, 1.5, 1.5, 'x', False, title)

        # y-Line Edge Case
        title = 'y-Line Edge Case 2'
        self.mask(0, 0.5, 1.5, 1.5, 'y', False, title)

    def test_roi_mask(self):

        # ROI Cut
        title = 'ROI Cut'
        self.mask(0, 0, 1.5, 1.5, 'roi', False, title)

        # ROI Cut Inverted
        title = 'ROI Cut Inverted'
        self.mask(0, 0, 1.5, 1.5, 'roi', True, title)

        # Moved ROI
        title = 'Moved ROI'
        self.mask(1, 1, 3, 1.5, 'roi', False, title)

        # ROI Cut Edge Case
        title = 'ROI Edge Case'
        self.mask(0, 0, 2, 1.5, 'roi', False, title)

        # ROI Cut Edge Case 2
        title = 'ROI Edge Case 2'
        self.mask(0, 0, 2.01, 1.5, 'roi', False, title)

    def test_border_mask(self):

        # Border Cut
        title = 'Border Cut'
        self.mask(0, 0, 1.5, 1.5, 'border', False, title)

        # Moved Border Cut
        title = 'Moved Border Cut'
        self.mask(1, 1, 2, 1.5, 'border', False, title)

        # Inverted Border Cut
        title = 'Inverted Border Cut'
        self.mask(0, 0, 1.5, 1.5, 'border', True, title)

        # Border Cut Edge Case
        title = 'Border Edge Case'
        self.mask(0, 0, 0.70, 1.5, 'border', False, title)

        # Border Cut Edge Case 2
        title = 'Border Edge Case 2'
        self.mask(0, 0, 2.13, 1.5, 'border', False, title)

    def test_outer_border_mask(self):

        # Outer Border Cut
        title = 'Outer Border Cut'
        self.mask(0, 0, 1.5, 1.5, 'outer_border', False, title)

        # Outer Border Edge Case
        title = 'Outer Border Edge Case'
        self.mask(0, 0, 0.00, 0.7, 'outer_border', False, title)

        # Outer Border Edge Case 2
        title = 'Outer Border Edge Case 2'
        self.mask(0, 0, 0.00, 2.13, 'outer_border', False, title)

    def test_ring_mask(self):

        # Ring Cut
        title = 'Ring Cut'
        self.mask(0, 0, 1.5, 1.5, 'ring', False, title)

        # Moved Ring Cut
        title = 'Moved Ring Cut'
        self.mask(1, 1, 2, 1.5, 'ring', False, title)

        # Inverted Ring Cut
        title = 'Inverted Ring Cut'
        self.mask(0, 0, 1.5, 1.5, 'ring', True, title)

        # Ring Edge Case 2
        title = 'Ring Edge Case'
        self.mask(0, 0, 1, 2.00, 'ring', False, title)

        # Ring Edge Case 2
        title = 'Ring Edge Case 2'
        self.mask(0, 0, 2.01, 1.5, 'ring', False, title)

    def mask(self, x, y, r, w, region, inverted, title):

        self.crosshair.set_position(x, y)
        self.crosshair.set_radius(r)
        self.crosshair.set_width(w)

        cut = self.crosshair.cut_from_data(
            self.data, region=region, inverted=inverted)

        expected_output = self.expected[title]

        npt.assert_equal(expected_output, cut.data)


if __name__ == '__main__':
    unittest.main()



# Python Imports
from abc import abstractmethod

# PyQt5 Imports
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QLabel, QCheckBox, QDoubleSpinBox, QWidget, QTabWidget


class CenteredLabel(QLabel):

    def __init__(self, *args, **kwargs):

        super(CenteredLabel, self).__init__(*args, **kwargs)
        self.setAlignment(Qt.AlignCenter)


class UseCheckBox(QCheckBox):

    def __init__(self, *args, **kwargs):

        super(UseCheckBox, self).__init__(*args, **kwargs)

        self.setChecked(True)


class AngleSpinBox(QDoubleSpinBox):

    def __init__(self, value, objectname, *args, **kwargs):

        super(AngleSpinBox, self).__init__(*args, **kwargs)

        self.setSuffix('°')
        self.setMinimum(-90)
        self.setMaximum(90)
        self.setValue(value)
        self.setDecimals(1)
        self.setSingleStep(1)
        self.setKeyboardTracking(False)
        self.setAlignment(Qt.AlignHCenter)
        self.setObjectName(objectname)


class WeightSpinBox(QDoubleSpinBox):

    def __init__(self, *args, value=1, **kwargs):

        super(WeightSpinBox, self).__init__(*args, **kwargs)

        self.setMinimum(0)
        self.setMaximum(99999.9)
        self.setValue(value)
        self.setDecimals(1)
        self.setSingleStep(0.1)
        self.setKeyboardTracking(False)
        self.setAlignment(Qt.AlignHCenter)
        self.setObjectName('weight')

class BackgroundSpinBox(QDoubleSpinBox):

    def __init__(self, *args, value=0, **kwargs):

        super(BackgroundSpinBox, self).__init__(*args, **kwargs)

        self.setMinimum(-99999.9)
        self.setMaximum(99999.9)
        self.setValue(value)
        self.setDecimals(1)
        self.setSingleStep(0.1)
        self.setKeyboardTracking(False)
        self.setAlignment(Qt.AlignHCenter)
        self.setObjectName('background')

class EnergySpinBox(QDoubleSpinBox):

    def __init__(self, *args, value=30, **kwargs):

        super(EnergySpinBox, self).__init__(*args, **kwargs)

        self.setSuffix('  eV')
        self.setMinimum(5)
        self.setMaximum(150)
        self.setValue(value)
        self.setDecimals(1)
        self.setSingleStep(0.1)
        self.setKeyboardTracking(False)
        self.setAlignment(Qt.AlignHCenter)
        self.setObjectName('energy')


class FixedSizeWidget(QWidget):

    def __init__(self, width, ratio, *args, **kwargs):

        super(FixedSizeWidget, self).__init__()

        height = width * ratio
        self.resize(width, height)
        self.setMaximumSize(width, height)
        self.setMinimumSize(width, height)


class AspectWidget(QWidget):

    def __init__(self, *args, ratio=0, **kwargs):

        self.ratio = ratio

        super().__init__(*args, **kwargs)

    def set_ratio(self, ratio):

        self.ratio = ratio

    def heightForWidth(self, width):

        return int(width / self.ratio)

    def widthForHeight(self, height):

        return int(height * self.ratio)

    def resizeEvent(self, event):
        '''BUG: For unknown reasons the resizeEvent is called multiple
        times with outdated sizes, thus if the user lowers the side that
        is too large, it will not update correctly.'''
        event.ignore()

        if self.ratio == 0:
            return

        old_width = event.oldSize().width()
        old_height = event.oldSize().height()
        new_width = event.size().width()
        new_height = event.size().height()

        if old_width == new_width:
            if old_height == new_height:
                # No change
                return

            else:
                # Only height changed -> use height
                height = new_height
                width = self.widthForHeight(height)

        else:
            if old_height == new_height:
                # Only width changed -> use width
                width = new_width
                height = self.heightForWidth(width)

            else:
                # Both changed -> use larger one
                if new_width >= new_height:
                    # Width is larger
                    width = new_width
                    height = self.heightForWidth(width)

                else:
                    # Height is larger
                    height = new_height
                    width = self.widthForHeight(height)

        self.blockSignals(True)
        self.resize(width, height)
        self.blockSignals(False)


class Tab(QTabWidget):

    close_requested = pyqtSignal()

    def __init__(self, *args, **kwargs):

        self.title = None

        super(Tab, self).__init__(*args, **kwargs)

    def get_title(self):

        return self.title

    def set_title(self, title):

        self.title = title

    def closeEvent(self, event):

        self.close_requested.emit()
        self.deleteLater()
        event.accept()
